import torch
import torch.nn as nn

from meloetta.frameworks.porygon.model import config
from meloetta.frameworks.porygon.model.utils import (
    _legal_policy,
    gather_along_rows,
)


class MaxMoveHead(nn.Module):
    def __init__(self, config: config.MaxMoveHeadConfig):
        super().__init__()

        self.key_fc = nn.Sequential(
            nn.Linear(config.entity_embedding_dim, config.entity_embedding_dim),
            nn.ReLU(),
            nn.Linear(config.entity_embedding_dim, config.key_dim),
        )
        self.query_fc = nn.Sequential(
            nn.Linear(config.autoregressive_embedding_dim, config.query_hidden_dim),
            nn.ReLU(),
            nn.Linear(config.query_hidden_dim, config.key_dim),
        )
        self.proj_move = nn.Sequential(
            nn.Linear(config.key_dim, config.entity_embedding_dim),
            nn.ReLU(),
            nn.Linear(config.entity_embedding_dim, config.autoregressive_embedding_dim),
        )

    def forward(
        self,
        action_type_index: torch.Tensor,
        autoregressive_embedding: torch.Tensor,
        max_moves: torch.Tensor,
        max_move_mask: torch.Tensor,
    ):
        T, B, *_ = autoregressive_embedding.shape

        query = self.query_fc(autoregressive_embedding)
        query = query.view(T, B, 1, 1, -1)
        keys = self.key_fc(max_moves)

        max_move_logits = query @ keys.transpose(-2, -1)
        max_move_logits = max_move_logits.view(T, B, -1)

        max_move_mask = max_move_mask.view(T * B, -1)
        max_move_mask[max_move_mask.sum() == 0] = True
        max_move_mask = max_move_mask.view(T, B, -1)

        max_move_policy = _legal_policy(max_move_logits, max_move_mask)
        embedding_index = torch.multinomial(max_move_policy.view(T * B, -1), 1)
        max_move_index = embedding_index.view(T, B, -1)

        move_embedding = gather_along_rows(keys.flatten(0, -3), embedding_index, 1)
        move_embedding = move_embedding.view(T, B, -1)
        projected_move_embedding = self.proj_move(move_embedding)

        valid_indices = action_type_index == 0
        valid_indices = valid_indices.unsqueeze(-1)

        autoregressive_embedding = (
            autoregressive_embedding + valid_indices * projected_move_embedding
        )

        return (
            max_move_logits,
            max_move_policy,
            max_move_index,
        )



import torch
import torch.nn as nn
import torch.nn.functional as F

from meloetta.frameworks.nash_ketchum.model.interfaces import EncoderOutput

from meloetta.frameworks.nash_ketchum.model import config
from meloetta.frameworks.nash_ketchum.model.utils import Resblock, VectorMerge


class Core(nn.Module):
    def __init__(self, config: config.CoreConfig):
        super().__init__()
        self.config = config

        self.merge = VectorMerge(
            {
                "pokemon_embedding": config.side_encoder_dim,
                "boosts": config.side_encoder_dim,
                "volatiles": config.side_encoder_dim,
                "side_conditions": config.side_encoder_dim,
                "weather_emb": config.side_encoder_dim,
                "scalar_emb": config.side_encoder_dim,
            },
            output_size=config.hidden_dim,
        )
        self.resblocks = nn.ModuleList(
            [Resblock(config.hidden_dim) for _ in range(config.num_layers)]
        )

    def forward(self, encoder_output: EncoderOutput):
        state_embedding = self.merge(
            {
                "pokemon_embedding": encoder_output.pokemon_embedding,
                "boosts": encoder_output.boosts,
                "volatiles": encoder_output.volatiles,
                "side_conditions": encoder_output.side_conditions,
                "weather_emb": encoder_output.weather_emb,
                "scalar_emb": encoder_output.scalar_emb,
            }
        )
        for resblock in self.resblocks:
            state_embedding = resblock(state_embedding)

        return state_embedding



import torch
import torch.nn as nn

from typing import Optional

from meloetta.frameworks.mewzero.model.interfaces import (
    State,
    EncoderOutput,
    Indices,
)
from meloetta.frameworks.mewzero.model.config import EncoderConfig
from meloetta.frameworks.mewzero.model.encoders import (
    PrivateEncoder,
    PrivateEncoderV2,
    PublicEncoder,
    PublicEncoderV2,
    ScalarEncoder,
    WeatherEncoder,
)


class Encoder(nn.Module):
    def __init__(self, gen: int, n_active: int, config: EncoderConfig):
        super().__init__()
        self.private_encoder = PrivateEncoderV2(
            gen=gen, n_active=n_active, config=config.private_encoder_config
        )
        self.public_encoder = PublicEncoderV2(
            gen=gen, n_active=n_active, config=config.public_encoder_config
        )
        self.weather_encoder = WeatherEncoder(
            gen=gen, config=config.weather_encoder_config
        )
        self.scalar_encoder = ScalarEncoder(
            gen=gen, n_active=n_active, config=config.scalar_encoder_config
        )

    def forward(self, state: State) -> EncoderOutput:
        # private info
        private_reserve = state["private_reserve"]

        # public info
        public_n = state["public_n"]
        public_total_pokemon = state["public_total_pokemon"]
        public_faint_counter = state["public_faint_counter"]
        public_side_conditions = state["public_side_conditions"]
        public_wisher = state["public_wisher"]
        public_active = state["public_active"]
        public_reserve = state["public_reserve"]
        public_stealthrock = state["public_stealthrock"]
        public_spikes = state["public_spikes"]
        public_toxicspikes = state["public_toxicspikes"]
        public_stickyweb = state["public_stickyweb"]

        # weather type stuff (still public)
        weather = state["weather"]
        weather_time_left = state["weather_time_left"]
        weather_min_time_left = state["weather_min_time_left"]
        pseudo_weather = state["pseudo_weather"]

        # scalar information
        turn = state["turn"]
        turns_since_last_move = state["turns_since_last_move"]
        prev_choices = state.get("prev_choices")
        choices_done = state.get("choices_done")

        # action masks
        action_type_mask = state["action_type_mask"]
        move_mask = state["move_mask"]
        switch_mask = state["switch_mask"]
        flag_mask = state["flag_mask"]
        max_move_mask = state.get("max_move_mask")
        target_mask = state.get("target_mask")

        (
            private_entity_emb,
            private_spatial,
            moves,
            switches,
        ) = self.private_encoder(private_reserve)

        public_entity_emb, public_spatial, = self.public_encoder(
            public_n,
            public_total_pokemon,
            public_faint_counter,
            public_side_conditions,
            public_wisher,
            public_active,
            public_reserve,
            public_stealthrock,
            public_spikes,
            public_toxicspikes,
            public_stickyweb,
        )

        weather_emb = self.weather_encoder(
            weather,
            weather_time_left,
            weather_min_time_left,
            pseudo_weather,
        )

        scalar_emb = self.scalar_encoder(
            turn,
            turns_since_last_move,
            prev_choices,
            choices_done,
            action_type_mask,
            move_mask,
            max_move_mask,
            switch_mask,
            flag_mask,
            target_mask,
        )

        return EncoderOutput(
            moves=moves,
            switches=switches,
            private_entity_emb=private_entity_emb,
            public_entity_emb=public_entity_emb,
            private_spatial=private_spatial,
            public_spatial=public_spatial,
            weather_emb=weather_emb,
            scalar_emb=scalar_emb,
        )



import math
import time
import wandb
import traceback
import threading

import torch
import torch.optim as optim
import torch.nn.functional as F

from typing import Tuple

from meloetta.frameworks.mewzero import utils
from meloetta.frameworks.mewzero.buffer import ReplayBuffer
from meloetta.frameworks.mewzero.config import MewZeroConfig
from meloetta.frameworks.mewzero.entropy import EntropySchedule
from meloetta.frameworks.mewzero.model import (
    MewZeroModel,
    TrainingOutput,
    Batch,
    Targets,
    Loss,
    Policy,
)
from meloetta.frameworks.mewzero.utils import (
    _player_others,
    v_trace,
    get_loss_nerd,
    get_loss_v,
)


class TargetNetSGD:
    def __init__(
        self,
        lr: float,
        target_model: MewZeroModel,
        learner_model: MewZeroModel,
    ):
        self.lr = lr
        self.target_model = target_model
        self.learner_model = learner_model
        self.param_groups = []

        for (name, param1), param2 in zip(
            self.target_model.named_parameters(), self.learner_model.parameters()
        ):
            if param2.requires_grad:
                self.param_groups.append({"params": param1, "name": name})

    def step(self):
        learner_state_dict = self.learner_model.state_dict()
        for group in self.param_groups:
            name = group["name"]
            param = group["params"]
            diff = learner_state_dict[name] - param.data
            param.data += self.lr * diff


class MewZeroLearner:
    def __init__(
        self,
        learner_model: MewZeroModel,
        actor_model: MewZeroModel,
        target_model: MewZeroModel,
        model_prev: MewZeroModel,
        model_prev_: MewZeroModel,
        replay_buffer: ReplayBuffer,
        config: MewZeroConfig = MewZeroConfig(),
    ):
        self.learner_model = learner_model
        self.actor_model = actor_model
        self.target_model = target_model
        self.model_prev = model_prev
        self.model_prev_ = model_prev_

        self.replay_buffer = replay_buffer

        self.config = config
        self._entropy_schedule = EntropySchedule(
            sizes=self.config.entropy_schedule_size,
            repeats=self.config.entropy_schedule_repeats,
        )
        self.optimizer = optim.Adam(
            self.learner_model.parameters(),
            lr=config.learning_rate,
            betas=(config.adam.b1, config.adam.b2),
            eps=config.adam.eps,
        )
        self.optimizer_target = TargetNetSGD(
            config.target_network_avg,
            self.target_model,
            self.learner_model,
        )
        self.learner_steps = 0

    def get_config(self):
        return {
            "parameters": self.learner_model.get_learnable_params(),
            "learner_steps": self.learner_steps,
            **self.config.__dict__,
        }

    def save(self):
        torch.save(
            {
                "learner_model": self.learner_model.state_dict(),
                "actor_model": self.actor_model.state_dict(),
                "target_model": self.target_model.state_dict(),
                "model_prev": self.model_prev.state_dict(),
                "model_prev_": self.model_prev_.state_dict(),
                "config": self.config,
                "optimizer": self.optimizer.state_dict(),
                "learner_steps": self.learner_steps,
            },
            f"cpkts/cpkt-{self.learner_steps:05}.tar",
        )

    @classmethod
    def from_config(self, config: MewZeroConfig = None):
        return MewZeroLearner.from_pretrained(config_override=config)

    @classmethod
    def from_pretrained(self, fpath: str = None, config_override: MewZeroConfig = None):
        assert not (
            fpath is None and config_override is None
        ), "please set one of `fpath` or `config`"

        if fpath is not None:
            datum = torch.load(fpath)
        else:
            datum = {}

        if datum.get("config"):
            print("Found config!")
            config: MewZeroConfig = datum["config"]

        if config_override is not None:
            config = config_override

        gen, gametype = utils.get_gen_and_gametype(config.battle_format)
        learner_model = MewZeroModel(
            gen=gen, gametype=gametype, config=config.model_config
        )
        learner_model.train()

        print(f"learnabled params: {learner_model.get_learnable_params():,}")

        actor_model = MewZeroModel(
            gen=gen, gametype=gametype, config=config.model_config
        )
        actor_model.load_state_dict(learner_model.state_dict())
        actor_model.share_memory()
        actor_model.eval()

        target_model = MewZeroModel(
            gen=gen, gametype=gametype, config=config.model_config
        )
        target_model.load_state_dict(learner_model.state_dict())
        target_model.eval()

        model_prev = MewZeroModel(
            gen=gen, gametype=gametype, config=config.model_config
        )
        model_prev.load_state_dict(learner_model.state_dict())
        model_prev.eval()

        model_prev_ = MewZeroModel(
            gen=gen, gametype=gametype, config=config.model_config
        )
        model_prev_.load_state_dict(learner_model.state_dict())
        model_prev_.eval()

        if datum.get("learner_model"):
            try:
                learner_model.load_state_dict(datum["learner_model"], strict=False)
                actor_model.load_state_dict(datum["actor_model"], strict=False)
                target_model.load_state_dict(datum["target_model"], strict=False)
                model_prev.load_state_dict(datum["model_prev"], strict=False)
                model_prev_.load_state_dict(datum["model_prev_"], strict=False)
            except Exception as e:
                traceback.print_exc()

        learner_model = learner_model.to(config.learner_device, non_blocking=True)
        actor_model = actor_model.to(config.actor_device, non_blocking=True)
        target_model = target_model.to(config.learner_device, non_blocking=True)
        model_prev = model_prev.to(config.learner_device, non_blocking=True)
        model_prev_ = model_prev_.to(config.learner_device, non_blocking=True)

        learner = MewZeroLearner(
            learner_model,
            actor_model,
            target_model,
            model_prev,
            model_prev_,
            replay_buffer=ReplayBuffer(
                config.trajectory_length,
                gen,
                gametype,
                config.num_buffers,
                config.learner_device,
            ),
            config=config,
        )

        if datum.get("optimizer"):
            learner.optimizer.load_state_dict(datum["optimizer"])
        else:
            print("Optimizer not loaded")

        if datum.get("learner_steps"):
            learner.learner_steps = datum["learner_steps"]
        else:
            print("Learner steps not loaded")

        return learner

    def collect_batch_trajectory(self):
        hidden_state = self.learner_model.core.initial_state(self.config.batch_size)
        hidden_state = (
            hidden_state[0].to(self.config.learner_device, non_blocking=True),
            hidden_state[1].to(self.config.learner_device, non_blocking=True),
        )
        batch = None
        while batch is None:
            time.sleep(1)
            batch = self.replay_buffer.get_batch(self.config.batch_size)
        return batch, hidden_state

    def step(self):
        batch, hidden_state = self.collect_batch_trajectory()
        alpha, update_target_net = self._entropy_schedule(self.learner_steps)
        self.update_paramters(batch, hidden_state, alpha, update_target_net)
        self.learner_steps += 1
        if self.learner_steps % 500 == 0:
            self.save()

    def update_paramters(
        self,
        batch: Batch,
        hidden_state: Tuple[torch.Tensor, torch.Tensor],
        alpha: float,
        update_target_net: bool,
        lock=threading.Lock(),
    ):
        with lock:
            self.optimizer.zero_grad(set_to_none=True)

            targets = self._get_targets(batch, hidden_state, alpha)
            losses = self._backpropagate(batch, hidden_state, targets)

            static_loss_dict = losses.to_log(batch)
            static_loss_dict["s"] = self.learner_steps
            wandb.log(static_loss_dict)

            torch.nn.utils.clip_grad_value_(
                self.learner_model.parameters(), self.config.clip_gradient
            )

            self.optimizer.step()
            with torch.no_grad():
                self.optimizer_target.step()

            if update_target_net:
                self.model_prev_.load_state_dict(self.model_prev.state_dict())
                self.model_prev.load_state_dict(self.learner_model.state_dict())

            self.actor_model.load_state_dict(self.learner_model.state_dict())

    def _get_targets(
        self,
        batch: Batch,
        hidden_state: Tuple[torch.Tensor, torch.Tensor],
        alpha: float,
    ) -> Targets:
        bidx = torch.arange(
            batch.batch_size, device=next(self.learner_model.parameters()).device
        )
        lengths = batch.valid.sum(0)
        final_reward = batch.rewards[lengths.squeeze() - 1, bidx]
        assert torch.all(final_reward.sum(-1) == 0).item(), "Env should be zero-sum!"

        learner: TrainingOutput
        target: TrainingOutput
        prev: TrainingOutput
        prev_: TrainingOutput

        with torch.no_grad():
            learner = self.learner_model.learning_forward(batch, hidden_state)
            target = self.target_model.learning_forward(batch, hidden_state)
            prev = self.model_prev.learning_forward(batch, hidden_state)
            prev_ = self.model_prev_.learning_forward(batch, hidden_state)

        is_vector = torch.unsqueeze(torch.ones_like(batch.valid), dim=-1)
        importance_sampling_corrections = [is_vector] * 2

        targets_dict = {
            "importance_sampling_corrections": importance_sampling_corrections
        }

        for policy_field, pi, log_pi, prev_log_pi, prev_log_pi_ in zip(
            *zip(*learner.pi._asdict().items()),
            learner.log_pi,
            prev.log_pi,
            prev_.log_pi,
        ):
            pi: torch.Tensor
            policy_field: str
            if pi is None:
                continue

            log_policy_reg = log_pi - (alpha * prev_log_pi + (1 - alpha) * prev_log_pi_)

            v_target_list, has_played_list, v_trace_policy_target_list = [], [], []

            for player in range(2):
                reward = batch.rewards[:, :, player]  # [T, B, Player]

                action_index = batch.get_index_from_policy(policy_field)
                action_oh = F.one_hot(action_index, pi.shape[-1])

                v_target_, has_played, policy_target_ = v_trace(
                    target.v,
                    batch.valid,
                    batch.player_id,
                    pi,
                    pi,
                    log_policy_reg,
                    _player_others(batch.player_id, batch.valid, player),
                    action_oh,
                    reward,
                    player,
                    lambda_=1.0,
                    c=self.config.c_vtrace,
                    rho=torch.inf,
                    eta=self.config.eta_reward_transform,
                    gamma=self.config.gamma,
                )
                v_target_list.append(v_target_)
                has_played_list.append(has_played)
                v_trace_policy_target_list.append(policy_target_)

            targets_dict[
                policy_field.replace("_policy", "_value_target")
            ] = v_target_list
            targets_dict[
                policy_field.replace("_policy", "_has_played")
            ] = has_played_list
            targets_dict[
                policy_field.replace("_policy", "_policy_target")
            ] = v_trace_policy_target_list

        return Targets(**targets_dict)

    def _backpropagate(
        self,
        batch: Batch,
        hidden_state: Tuple[torch.Tensor, torch.Tensor],
        targets: Targets,
    ) -> Loss:

        loss_dict = {key + "_loss": 0 for key in Policy._fields}
        loss_dict.update({key.replace("policy", "value"): 0 for key in loss_dict})

        fields = (
            "action_type",
            "move",
            "switch",
        )
        if self.actor_model.gen >= 6:
            fields += ("flag",)

        if self.actor_model.gen == 8:
            fields += ("max_move",)

        if self.actor_model.gametype != "singles":
            fields += ("target",)

        global_action_type = batch.action_type_index

        total_valid = batch.valid.sum().item()

        total_move_valid = batch.valid * (global_action_type == 0)
        total_move_valid *= batch.move_mask.sum(-1) > 1
        total_move_valid = total_move_valid.sum().item()

        total_switch_valid = batch.valid * (global_action_type == 1)
        total_switch_valid *= batch.switch_mask.sum(-1) > 1
        total_switch_valid = total_switch_valid.sum().item()

        if batch.flag_mask is not None:
            total_flag_valid = batch.valid * (global_action_type == 0)
            total_flag_valid *= batch.flag_mask.sum(-1) > 1
            total_flag_valid = total_flag_valid.sum().item()

        minibatch_size = min(128, (4096 // batch.trajectory_length) + 1)

        for batch_index in range(math.ceil(batch.batch_size / minibatch_size)):
            mini_state_h, mini_state_c = hidden_state

            batch_start = minibatch_size * batch_index
            batch_end = minibatch_size * (batch_index + 1)

            mini_state_h = mini_state_h[:, batch_start:batch_end].contiguous()
            mini_state_c = mini_state_c[:, batch_start:batch_end].contiguous()

            loss = 0

            max_length = batch.valid[:, batch_start:batch_end].sum(0).max()
            minibatch = batch.slice(batch_start, batch_end, max_length=max_length)
            targ = targets.slice(batch_start, batch_end, max_length=max_length)

            output, (mini_state_c, mini_state_h) = self.learner_model.forward(
                minibatch._asdict(),
                (mini_state_c, mini_state_h),
                training=True,
                need_log_policy=False,
            )

            for field in fields:
                pi_field = field + "_policy"
                logit_field = field + "_logits"

                pi = getattr(output.pi, pi_field)
                logit = getattr(output.logit, logit_field)

                if pi is None:
                    continue

                action_type_index = minibatch.action_type_index
                policy_mask = minibatch.get_mask_from_policy(pi_field)

                if field == "action_type":
                    valid = torch.ones_like(action_type_index)
                elif field == "switch":
                    valid = action_type_index == 1
                elif field == "target":
                    valid = action_type_index == 2
                else:
                    valid = action_type_index == 0

                valid *= policy_mask.sum(-1) > 1

                value_loss = get_loss_v(
                    [output.v] * 2,
                    targ.get_value_target(pi_field),
                    targ.get_has_played(pi_field, valid),
                )

                # Uses v-trace to define q-values for Nerd
                policy_loss = get_loss_nerd(
                    [logit] * 2,
                    [pi] * 2,
                    targ.get_policy_target(pi_field),
                    minibatch.valid * valid,
                    minibatch.player_id,
                    policy_mask,
                    targ.importance_sampling_corrections,
                    clip=self.config.nerd.clip,
                    threshold=self.config.nerd.beta,
                )

                policy_loss_field = field + "_policy_loss"
                value_loss_field = field + "_value_loss"

                loss_dict[policy_loss_field] += policy_loss.item()
                loss_dict[value_loss_field] += value_loss.item()

                if "action_type" in field:
                    policy_valid = total_valid
                elif "move" in field:
                    policy_valid = total_move_valid
                elif "flag" in field:
                    policy_valid = total_flag_valid
                elif "switch" in field:
                    policy_valid = total_switch_valid

                loss += (value_loss + policy_loss) / policy_valid

            loss.backward()

        for key in loss_dict:
            if "action_type" in key:
                loss_dict[key] /= total_valid
            elif "move" in key:
                loss_dict[key] /= total_move_valid
            elif "flag" in key:
                loss_dict[key] /= total_flag_valid
            elif "switch" in key:
                loss_dict[key] /= total_switch_valid

        return Loss(**loss_dict)

    def run(self):
        while True:
            self.step()



import torch
import multiprocessing as mp

from meloetta.room import BattleRoom
from meloetta.actors.base import Actor
from meloetta.actors.types import State, Choices

from meloetta.frameworks.max_damage.model import MaxDamageModel


class MaxDamageActor(Actor):
    def __init__(self, gen: int = 9, queue: mp.Queue = None):
        self.model = MaxDamageModel(gen=gen)
        self.model.eval()

        self.queue = queue

    def choose_action(
        self,
        state: State,
        room: BattleRoom,
        choices: Choices,
    ):
        with torch.no_grad():
            func, args, kwargs = self.model(state, choices)

        return func, args, kwargs

    def post_match(self, room: BattleRoom):
        if self.queue is not None:
            datum = room.get_reward()
            reward = datum["reward"]

            if reward > 0:
                self.queue.put(("maxdmg", 1))
            elif reward < 0:
                self.queue.put(("maxdmg", 0))



import torch.nn as nn

from copy import deepcopy
from typing import Optional, Union, Literal, Dict, Any

from meloetta.actors.types import State

from meloetta.frameworks.nash_ketchum.model.utils import _log_policy, MLP

from meloetta.frameworks.nash_ketchum.model import config
from meloetta.frameworks.nash_ketchum.model import (
    Encoder,
    PolicyHeads,
    Core,
    Policy,
    ValueHead,
    Indices,
    ModelOutput,
    PostProcess,
)

from meloetta.data import _STATE_FIELDS


class NAshKetchumModel(nn.Module):
    def __init__(
        self,
        gen: int = 9,
        gametype: Literal["singles", "doubles", "triples"] = "singles",
        config: config.NAshKetchumModelConfig = config.NAshKetchumModelConfig(),
    ) -> None:
        super().__init__()

        self.gen = gen
        self.gametype = gametype
        n_active = {"singles": 1, "doubles": 2, "triples": 3}[gametype]

        self.encoder = Encoder(gen=gen, n_active=n_active, config=config.encoder_config)
        self.policy_heads = PolicyHeads(
            gen=gen, gametype=gametype, config=config.policy_heads_config
        )

        self.state_fields = deepcopy(_STATE_FIELDS)

        if gen != 8:
            self.state_fields.remove("max_move_mask")

        if self.gametype == "singles":
            self.state_fields.remove("target_mask")
            self.state_fields.remove("prev_choices")

        self.core = Core(config.resnet_core_config)
        self.value_Head = ValueHead(config.value_head_config)

        hidden_dim = config.resnet_core_config.hidden_dim
        self.state_action_mlp = MLP([hidden_dim, hidden_dim, hidden_dim])
        self.state_mlp = MLP([hidden_dim, hidden_dim, hidden_dim])

    def get_learnable_params(self) -> int:
        return sum([p.numel() for p in self.parameters() if p.requires_grad])

    def clean(self, state: State) -> State:
        return {k: state[k] for k in self.state_fields}

    def forward(
        self, state: State, compute_value: bool = True, indices: Indices = None
    ) -> ModelOutput:
        encoder_output = self.encoder.forward(state)
        state_emb = self.core.forward(encoder_output)
        if compute_value:
            value = self.value_Head.forward(state_emb)
        else:
            value = None
        indices, logits, policy, _ = self.policy_heads.forward(
            state_emb,
            encoder_output,
            state,
            indices,
        )
        return ModelOutput(
            indices=indices,
            logit=logits,
            pi=policy,
            v=value,
            # state_action_emb=self.state_action_mlp(state_action_emb),
            # state_emb=self.state_mlp(state_emb),
        )

    def acting_forward(self, state: State) -> ModelOutput:
        return self.forward(state, compute_value=False)

    def learning_forward(self, state: State, indices: Indices = None) -> ModelOutput:
        model_output = self.forward(state, compute_value=True, indices=indices)
        return self.postprocess(state, model_output)

    def postprocess(
        self,
        state: State,
        model_output: ModelOutput,
        choices: Optional[Dict[str, Any]] = None,
    ) -> Union[ModelOutput, PostProcess]:
        """post process method"""

        if choices:
            targeting = bool(state.get("targeting"))
            output = self.index_to_action(targeting, model_output.indices, choices)

        else:
            output = ModelOutput(
                indices=model_output.indices,
                pi=model_output.pi,
                v=model_output.v,
                log_pi=self.get_log_policy(model_output.logit, state),
                logit=model_output.logit,
            )

        return output

    def get_log_policy(self, logits: Policy, state: State):
        if self.gametype != "singles":
            target_log_policy = _log_policy(
                logits.target,
                state["target_mask"],
            )
        else:
            target_log_policy = None

        if self.gen == 8:
            max_move_log_policy = _log_policy(
                logits.max_move,
                state["max_move_mask"],
            )
        else:
            max_move_log_policy = None

        if self.gen >= 6:
            flag_log_policy = _log_policy(
                logits.flag,
                state["flag_mask"],
            )
        else:
            flag_log_policy = None

        return Policy(
            action_type=_log_policy(logits.action_type, state["action_type_mask"]),
            move=_log_policy(logits.move, state["move_mask"]),
            switch=_log_policy(logits.switch, state["switch_mask"]),
            max_move=max_move_log_policy,
            flag=flag_log_policy,
            target=target_log_policy,
        )

    def index_to_action(
        self,
        targeting: bool,
        indices: Indices,
        choices: Optional[Dict[str, Any]] = None,
    ):
        action_type_index = indices.action_type_index.item()
        move_index = indices.move_index.item()
        switch_index = indices.switch_index.item()

        if not targeting:
            index = action_type_index

            if action_type_index == 0:
                flag_index = (indices.flag_index).item()
                if flag_index == 3:
                    max_move_index = (indices.max_move_index).item()
                    index = max_move_index
                else:
                    index = move_index

                if choices is not None:
                    if flag_index == 0:
                        data = choices["moves"]
                    elif flag_index == 1:
                        data = choices["mega_moves"]
                    elif flag_index == 2:
                        data = choices["zmoves"]
                    elif flag_index == 3:
                        data = choices["max_moves"]
                    elif flag_index == 4:
                        data = choices["tera_moves"]
                    else:
                        data = choices["moves"]
                else:
                    data = None

            elif action_type_index == 1:
                if choices is not None:
                    data = choices["switches"]
                else:
                    data = None
                index = switch_index

        else:
            if choices is not None:
                data = choices["targets"]
            else:
                data = None

            target_index = (indices.target_index).item()
            index = target_index

        try:
            return PostProcess(data, index)
        except:
            print()



import torch
import torch.nn as nn

from meloetta.frameworks.nash_ketchum.model import config
from meloetta.frameworks.nash_ketchum.model.utils import (
    Resblock,
    _legal_policy,
    _multinomial,
    BIAS,
)


class ActionTypeHead(nn.Module):
    def __init__(self, config: config.ActionTypeHeadConfig):
        super().__init__()

        self.resblocks = nn.Sequential(
            *[Resblock(config.residual_dim) for _ in range(1)]
        )
        self.action_fc = nn.Linear(config.residual_dim, config.num_action_types)
        self.action_type_emb = nn.Embedding(
            config.num_action_types, config.state_embedding_dim
        )

    def forward(
        self,
        state: torch.Tensor,
        action_type_mask: torch.Tensor,
        action_type_index: torch.Tensor = None,
    ):
        T, B, *_ = state.shape

        embedded_state = self.resblocks(state)
        action_type_logits = self.action_fc(embedded_state)

        action_type_policy = _legal_policy(action_type_logits, action_type_mask)
        action_type_logits = torch.where(action_type_mask, action_type_logits, BIAS)
        action_type_policy = action_type_policy.view(T * B, -1)

        if action_type_index is None:
            action_type_index = _multinomial(action_type_policy)

        action_type_policy = action_type_policy.view(T, B, -1)
        action_type_index = action_type_index.view(T, B)

        action_type_embedding = self.action_type_emb(action_type_index.long())
        autoregressive_embedding = embedded_state + action_type_embedding

        return (
            action_type_logits,
            action_type_policy,
            action_type_index,
            autoregressive_embedding,
        )



import json
import time
import torch
import traceback

from typing import Dict, Any

from torch.nn.utils.rnn import pad_sequence

from meloetta.room import BattleRoom

from meloetta.frameworks.nash_ketchum import ReplayBuffer
from meloetta.frameworks.nash_ketchum.model import (
    NAshKetchumModel,
    ModelOutput,
    PostProcess,
)


from meloetta.actors.base import Actor
from meloetta.actors.types import State, Choices, Battle
from meloetta.utils import expand_bt
from meloetta.data import (
    BOOSTS,
    VOLATILES,
    PSEUDOWEATHERS,
    SIDE_CONDITIONS,
    get_item_effect_token,
    get_status_token,
    get_gender_token,
    get_species_token,
    get_ability_token,
    get_item_token,
    get_move_token,
    get_type_token,
    get_weather_token,
)


SIDES = ["mySide", "farSide"]


class NAshKetchumActor(Actor):
    def __init__(
        self,
        model: NAshKetchumModel,
        replay_buffer: ReplayBuffer = None,
        pid: str = None,
    ):
        self.model = model
        self.gen = model.gen

        self.replay_buffer = replay_buffer

        self.step_index = 0
        self.index = None

        self.env_outputs = []
        self.model_outputs = []

        self.pid = pid

    @property
    def storing_transition(self):
        return self.replay_buffer is not None

    @torch.no_grad()
    def choose_action(
        self,
        state: State,
        room: BattleRoom,
        choices: Choices,
    ):
        model_output: ModelOutput
        postprocess: PostProcess

        try:
            model_output = self.model.acting_forward(state)
            postprocess = self.model.postprocess(
                state=state,
                model_output=model_output,
                choices=choices,
            )

            data = postprocess.data
            index = postprocess.index
            func, args, kwargs = data[index]
        except:
            trace = traceback.format_exc()
            print(trace)

            battle_tag = room.battle_tag

            json_state = room.get_state()
            side = json_state["side"]

            datum = {
                "json_state": json_state,
                "choices": {
                    k: {sk: sv[1:] for sk, sv in v.items()} for k, v in choices.items()
                },
                "indices": model_output.indices.to_json(),
                "policy": model_output.pi.to_json(),
                "logits": model_output.logit.to_json(),
                "traceback": trace.split("\n"),
                "state_dict": self.model.state_dict(),
                "state": state,
            }
            torch.save(datum, f"errors/{battle_tag}-{side}-model.pt")

            return self.choose_action(state, room, choices)
        else:
            if self.storing_transition:
                self.store_transition(state, model_output, room)

        return func, args, kwargs

    def get_index(self, battle_tag: str):
        if self.index is None:
            self.index = self.replay_buffer._get_index(battle_tag)
        return self.index

    def store_transition(
        self, state: State, model_output: ModelOutput, room: BattleRoom
    ):
        state = self.model.clean(state)
        to_store = model_output.to_store(state)
        index = self.get_index(room.battle_tag)
        self.replay_buffer.store_sample(index, to_store, self.pid)

    def post_match(self, room: BattleRoom):
        if self.storing_transition:
            datum = room.get_reward()
            index = self.get_index(room.battle_tag)

            self.replay_buffer.append_reward(index, self.pid, datum["reward"])
            self.replay_buffer.register_done(room.battle_tag)

    def _vectorize_pokemon(
        self, pokemon: Dict[str, Any], public: bool = False
    ) -> torch.Tensor:

        if public:
            moves = {}
        else:
            moves = {
                get_move_token(self.gen, "id", move): -1 for move in pokemon["moves"]
            }

        for move, pp in pokemon.get("moveTrack", []):
            token = get_move_token(self.gen, "id", move)
            moves[token] = pp

        move_keys = list(moves.keys())[:4]
        move_values = list(moves.values())[:4]

        move_keys += [-1 for _ in range(4 - len(move_keys))]
        move_values += [-1 for _ in range(4 - len(move_values))]

        hp = pokemon.get("hp", 0) * (not public)
        maxhp = pokemon.get("maxhp", 0) * (not public)
        hp_ratio = pokemon.get("hp", 0) / max(pokemon.get("maxhp", 0), 1)
        fainted = hp_ratio == 0

        last_move_token = get_move_token(self.gen, "id", pokemon.get("lastMove"))

        data = [
            get_species_token(self.gen, "name", pokemon["name"]),
            get_species_token(self.gen, "forme", pokemon["speciesForme"]),
            pokemon.get("slot", 0),
            hp,
            maxhp,
            hp_ratio,
            pokemon.get("stats", {}).get("atk", -1),
            pokemon.get("stats", {}).get("def", -1),
            pokemon.get("stats", {}).get("spa", -1),
            pokemon.get("stats", {}).get("spd", -1),
            pokemon.get("stats", {}).get("spe", -1),
            fainted,
            pokemon.get("active", 0),
            pokemon.get("level", 1),
            get_gender_token(pokemon.get("gender", "")),
            get_ability_token(self.gen, "name", pokemon["ability"]),
            get_ability_token(self.gen, "name", pokemon["baseAbility"]),
            get_item_token(self.gen, "name", pokemon.get("item", "")),
            get_item_token(self.gen, "name", pokemon.get("prevItem", "")),
            get_item_effect_token(pokemon.get("itemEffect", "")),
            get_item_effect_token(pokemon.get("prevItemEffect", "")),
            get_status_token(pokemon.get("status", "")),
            min(pokemon.get("statusData", {}).get("sleepTurns", 0), 3),
            min(pokemon.get("statusData", {}).get("toxicTurns", 0), 15),
            last_move_token,
            moves.get(last_move_token, 0),
            *move_keys,
            *move_values,
            get_type_token(self.gen, pokemon.get("terastallized")),
            get_type_token(self.gen, pokemon.get("teraType")),
            min(pokemon.get("timesAttacked", 0), 6),
        ]
        return torch.tensor(data)

    def get_vectorized_state(
        self, room: BattleRoom, battle: Battle
    ) -> Dict[str, torch.Tensor]:
        my_private_side = {p["searchid"]: p for p in battle["myPokemon"]}

        my_public_side = {
            p["searchid"]: p for p in battle["mySide"]["pokemon"] if p is not None
        }
        my_active = {p["searchid"] for p in battle["mySide"]["active"] if p is not None}

        opp_public_side = {
            p["searchid"]: p for p in battle["farSide"]["pokemon"] if p is not None
        }
        opp_active = {
            p["searchid"] for p in battle["farSide"]["active"] if p is not None
        }

        boosts = [None, None]
        volatiles = [None, None]

        for sid, (public_side, active_list) in enumerate(
            zip([my_public_side, opp_public_side], [my_active, opp_active])
        ):
            for pokemon in public_side.values():
                if pokemon["searchid"] in active_list:
                    boosts_vector = torch.tensor(
                        [pokemon["boosts"].get(boost, 0) for boost in BOOSTS]
                    )
                    boosts[sid] = boosts_vector

                    volatile_vector = torch.tensor(
                        [
                            1 if volatile in pokemon["volatiles"] else 0
                            for volatile in VOLATILES
                        ]
                    )
                    volatiles[sid] = volatile_vector

        for sid in range(len(boosts)):
            if boosts[sid] is None:
                boosts[sid] = torch.zeros(len(BOOSTS), dtype=torch.long)

        for sid in range(len(volatiles)):
            if volatiles[sid] is None:
                volatiles[sid] = torch.zeros(len(VOLATILES), dtype=torch.long)

        boosts = torch.stack(boosts)
        volatiles = torch.stack(volatiles)

        side_conditions = []
        n = []
        total_pokemon = []
        faint_counter = []
        wisher_slot = []

        for side in SIDES:
            side_datum = battle[side]
            sc_datum = battle[side]["sideConditions"]

            side_conditions_vector = [
                sc_datum.get(side_condition, [None, 0, 0, 0])[1:]
                for side_condition in SIDE_CONDITIONS
                if side_condition
                not in {"stealthrock", "spikes", "toxicspikes", "stickyweb"}
            ]
            side_conditions_vector = [i for o in side_conditions_vector for i in o]
            side_conditions_vector += [sc_datum.get("stealthrock", [None, 0])[1]]
            side_conditions_vector += [sc_datum.get("toxicspikes", [None, 0])[1]]
            side_conditions_vector += [sc_datum.get("spikes", [None, 0])[1]]
            side_conditions_vector += [sc_datum.get("stickyweb", [None, 0])[1]]
            side_conditions += [torch.tensor(side_conditions_vector)]

            n += [torch.tensor(side_datum["n"])]
            total_pokemon += [torch.tensor(side_datum["totalPokemon"])]
            faint_counter += [torch.tensor(min(6, side_datum["faintCounter"]))]
            wisher_slot += [
                torch.tensor((side_datum["wisher"] or {"slot": -1})["slot"])
            ]

        side_conditions = torch.stack(side_conditions)
        n = torch.stack(n)
        total_pokemon = torch.stack(total_pokemon)
        faint_counter = torch.stack(faint_counter)
        wisher_slot = torch.stack(wisher_slot)

        my_private_side_vectors = []
        for i, (sid, private_data) in enumerate(my_private_side.items()):
            datum = {}

            public_data = my_public_side.get(sid, {})
            if i == 0:
                keys = list(public_data) + list(private_data)

            datum["active"] = sid in my_active
            for key in keys:
                value = private_data.get(key) or public_data.get(key)
                if value is not None:
                    datum[key] = value

            private_vector = self._vectorize_pokemon(datum)
            my_private_side_vectors.append(private_vector)

        my_public_side_vectors = []
        for i, (sid, public_data) in enumerate(my_public_side.items()):
            datum = {}
            datum["active"] = sid in my_active
            for key in keys:
                if key in public_data:
                    value = public_data.get(key)
                    if value is not None:
                        datum[key] = value

            public_vector = self._vectorize_pokemon(datum, public=True)
            my_public_side_vectors.append(public_vector)

        opp_public_side_vectors = []
        for i, (sid, public_data) in enumerate(opp_public_side.items()):
            datum = {}
            datum["active"] = sid in opp_active
            for key in keys:
                if key in public_data:
                    value = public_data.get(key)
                    if value is not None:
                        datum[key] = value

            public_vector = self._vectorize_pokemon(datum, public=True)
            opp_public_side_vectors.append(public_vector)

        my_private_side = torch.stack(my_private_side_vectors)
        my_private_side = torch.cat(
            (my_private_side, -torch.ones_like(my_private_side))
        )

        my_public_side = torch.stack(my_public_side_vectors)
        opp_public_side = torch.stack(opp_public_side_vectors)

        sides = pad_sequence(
            [my_private_side, my_public_side, opp_public_side],
            batch_first=True,
            padding_value=-1,
        )

        pseudoweathers = {
            (pseudoweather[0]): pseudoweather[1:]
            for pseudoweather in battle["pseudoWeather"]
        }
        pseudoweathers = [
            torch.tensor(pseudoweathers.get(pseudoweather, [-1, -1]))
            for pseudoweather in PSEUDOWEATHERS
        ]
        pseudoweathers = torch.stack(pseudoweathers)

        weather = torch.tensor(
            [
                get_weather_token(battle["weather"]),
                battle["weatherTimeLeft"],
                battle["weatherMinTimeLeft"],
            ]
        )

        turn = torch.tensor([battle["turn"]])
        scalars = torch.cat((turn, n, total_pokemon, faint_counter), dim=-1).long()

        state = {
            "sides": sides,
            "boosts": boosts,
            "volatiles": volatiles,
            "weather": weather,
            "side_conditions": side_conditions,
            "pseudoweathers": pseudoweathers,
            "wisher": wisher_slot,
            "scalars": scalars,
        }
        state = {key: expand_bt(value) for key, value in state.items()}
        return state



import math

import numpy as np

import torch
import torch.nn as nn
import torch.nn.functional as F

from typing import Optional, Sequence, Mapping, Dict

BIAS = -1e4


def binary_enc_matrix(num_embeddings: int):
    bits = math.ceil(math.log2(num_embeddings))
    mask = 2 ** torch.arange(bits)
    x = torch.arange(mask.sum().item() + 1)
    embs = x.unsqueeze(-1).bitwise_and(mask).ne(0).float()
    return embs[..., :num_embeddings:, :]


def sqrt_one_hot_matrix(num_embeddings: int):
    x = torch.arange(num_embeddings)
    i = (x**0.5).to(torch.long)
    return F.one_hot(i, torch.max(i) + 1).to(torch.float)


def power_one_hot_matrix(num_embeddings: int, power: float):
    x = torch.arange(num_embeddings)
    i = (x**power).to(torch.long)
    return F.one_hot(i, torch.max(i) + 1).to(torch.float)


def _legal_policy(logits: torch.Tensor, legal_actions: torch.Tensor) -> torch.Tensor:
    """A soft-max policy that respects legal_actions."""
    # Fiddle a bit to make sure we don't generate NaNs or Inf in the middle.
    masked_logits = torch.where(legal_actions, logits, float("-inf"))
    return masked_logits.softmax(-1)


def _multinomial(dist: torch.Tensor) -> torch.Tensor:
    """
    We use this method instead of `torch.multinomial` for two reasons:
    see: https://github.com/pytorch/pytorch/blob/9e089db32e6274e8eb094829afc97f2ddcd6f1ef/aten/src/ATen/native/Distributions.cpp#L632

    - The current version of torch is bugged for multinomial and selects invalid indices with a very small prob.
    - numpy is significantly faster at the operation, even on the cpu
    """
    np_dist = dist.cpu().view(-1).numpy().astype(np.float64)
    np_dist = np_dist / sum(np_dist)
    index = np.random.multinomial(1, np_dist)
    index = index.argmax(-1)
    index = torch.tensor(index)
    return index


def _log_policy(logits: torch.Tensor, legal_actions: torch.Tensor) -> torch.Tensor:
    log_policy = torch.where(legal_actions, logits, float("-inf"))
    log_policy = log_policy.log_softmax(-1)
    log_policy = torch.masked_fill(log_policy, ~legal_actions, 0)
    return log_policy


def gather_along_rows(
    x: torch.Tensor, index: torch.Tensor, dim: int = 1, keepdim: bool = True
):
    """
    x: torch.Tensor -> [B, T, D]
    dim: int
    index: torch.Tensor -> [B,], values in range (0, T)
    """
    assert len(x.shape) == 3
    B, _, D = x.shape
    index = index.view(B, 1).repeat(1, D).view(B, 1, D)
    rows = torch.gather(x, dim, index)
    if keepdim:
        return rows
    else:
        return rows.squeeze(dim)


class MLP(nn.Module):
    def __init__(self, layer_sizes: Sequence[int], use_layer_norm: bool = True):
        super().__init__()
        self.mlp = nn.Sequential(
            *[
                nn.Sequential(
                    *(
                        ([nn.LayerNorm(in_features)] if use_layer_norm else [])
                        + [nn.ReLU(), nn.Linear(in_features, out_features)]
                    )
                )
                for in_features, out_features, in zip(layer_sizes, layer_sizes[1:])
            ]
        )

    def forward(self, x: torch.Tensor):
        return self.mlp(x)


class GLU(nn.Module):
    def __init__(self, input_dim: int, output_dim: int, context_dim: int):
        super().__init__()
        self.lin1 = MLP([context_dim, input_dim])
        self.lin2 = MLP([input_dim, output_dim])

    def forward(self, x: torch.Tensor, context: torch.Tensor):
        gate = self.lin1(context)
        gate = torch.sigmoid(gate)
        x = gate * x
        x = self.lin2(x)
        return x


class Resblock(nn.Module):
    def __init__(
        self,
        hidden_size: int,
        num_layers: int = 2,
        use_layer_norm: bool = True,
    ) -> None:
        super().__init__()

        layers = []
        for _ in range(num_layers):
            hiddens = [nn.ReLU(), nn.Linear(hidden_size, hidden_size)]
            if use_layer_norm:
                hiddens = [nn.LayerNorm(hidden_size)] + hiddens
            layers.append(nn.Sequential(*hiddens))
        self.layers = nn.ModuleList(layers)

    def forward(self, x: torch.Tensor):
        shortcut = x
        for layer in self.layers:
            x = layer(x)
        return x + shortcut


class MultiHeadAttention(nn.Module):
    def __init__(
        self,
        input_size: int,
        num_heads: int,
        key_size: int,
        value_size: int = None,
        model_size: int = None,
    ):
        super().__init__()

        self.num_heads = num_heads
        self.key_size = key_size
        value_size = value_size or key_size
        model_size = model_size or key_size * num_heads
        self.value_size = value_size
        self.model_size = model_size

        self.lin_q = nn.Linear(input_size, num_heads * key_size)
        self.lin_k = nn.Linear(input_size, num_heads * key_size)
        self.lin_v = nn.Linear(input_size, num_heads * value_size)
        self.lin_out = nn.Linear(num_heads * value_size, model_size)

        self.attn_denom = key_size**0.5

    def _linear_projection(
        self, x: torch.Tensor, layer: nn.Module, head_size: int
    ) -> torch.Tensor:
        y = layer(x)
        *leading_dims, _ = x.shape
        return y.reshape((*leading_dims, self.num_heads, head_size))

    def forward(
        self,
        query: torch.Tensor,
        key: torch.Tensor,
        value: torch.Tensor,
        mask: torch.Tensor,
    ):
        *leading_dims, sequence_length, _ = query.shape

        query_heads = self._linear_projection(query, self.lin_q, self.key_size)
        key_heads = self._linear_projection(key, self.lin_k, self.key_size)
        value_heads = self._linear_projection(value, self.lin_v, self.value_size)

        attn_logits = torch.einsum("...thd,...Thd->...htT", query_heads, key_heads)
        attn_logits = attn_logits / self.attn_denom

        if mask is not None:
            if mask.ndim == 2:
                mask = mask.view(*leading_dims, 1, 1, sequence_length)
                mask = mask.expand(-1, self.num_heads, -1, -1)
            if mask.ndim != attn_logits.ndim:
                raise ValueError(
                    f"Mask dimensionality {mask.ndim} must match logits dimensionality "
                    f"{attn_logits.ndim}."
                )
            attn_logits = torch.where(mask, attn_logits, float("-inf"))
        attn_weights = torch.softmax(attn_logits, dim=-1)  # [H, T', T]

        # Weight the values by the attention and flatten the head vectors.
        attn = torch.einsum("...htT,...Thd->...thd", attn_weights, value_heads)
        attn = torch.reshape(attn, (*leading_dims, sequence_length, -1))  # [T', H*V]

        # Apply another projection to get the final embeddings.
        return self.lin_out(attn)  # [T', D']


class TransformerLayer(nn.Module):
    def __init__(
        self,
        transformer_num_heads: int,
        transformer_key_size: int,
        transformer_value_size: int,
        input_size: int,
        model_size: int,
        use_layer_norm: bool = True,
    ) -> None:
        super().__init__()
        self._use_layer_norm = use_layer_norm
        if use_layer_norm:
            self.ln = nn.LayerNorm(model_size)
        self.mha = MultiHeadAttention(
            num_heads=transformer_num_heads,
            key_size=transformer_key_size,
            value_size=transformer_value_size,
            model_size=model_size,
            input_size=input_size,
        )

    def forward(self, x: torch.Tensor, mask: torch.Tensor):
        x1 = x
        if self._use_layer_norm:
            x1 = self.ln(x1)

        x1 = F.relu(x1)
        # The logits mask has shape [num_heads, num_units, num_units]:
        x1 = self.mha(x1, x1, x1, mask)

        # Mask here mostly for safety:
        x1 = torch.where(mask.unsqueeze(-1), x1, 0)
        x = x + x1

        return x


class TransformerEncoder(nn.Module):
    def __init__(
        self,
        model_size: int,
        num_layers: int,
        num_heads: int,
        key_size: int,
        value_size: int,
        resblocks_num_before: int,
        resblocks_num_after: int,
        resblocks_hidden_size: Optional[int] = None,
        use_layer_norm: bool = True,
    ):

        super().__init__()
        self._transformer_num_layers = num_layers
        self._transformer_num_heads = num_heads
        self._transformer_key_size = key_size
        self._transformer_value_size = value_size

        if resblocks_hidden_size is None:
            resblocks_hidden_size = model_size

        self._transformer_layers = nn.ModuleList(
            [
                TransformerLayer(
                    num_heads,
                    key_size,
                    value_size,
                    input_size=resblocks_hidden_size,
                    model_size=model_size,
                    use_layer_norm=use_layer_norm,
                )
                for _ in range(num_layers)
            ]
        )

        self._resblocks_before = nn.ModuleList(
            [
                Resblock(
                    hidden_size=resblocks_hidden_size,
                    use_layer_norm=use_layer_norm,
                )
                for _ in range(resblocks_num_before)
            ]
        )
        self._resblocks_after = nn.ModuleList(
            [
                Resblock(
                    hidden_size=resblocks_hidden_size,
                    use_layer_norm=use_layer_norm,
                )
                for _ in range(resblocks_num_after)
            ]
        )

        self._resblocks_hidden_size = resblocks_hidden_size
        self._use_layer_norm = use_layer_norm

    def forward(self, x: torch.Tensor, mask: torch.Tensor = None) -> torch.Tensor:
        for resblock_before in self._resblocks_before:
            x = resblock_before(x)

        for transformer_layer in self._transformer_layers:
            x = transformer_layer(x, mask)

        for resblock_after in self._resblocks_after:
            x = resblock_after(x)

        # Mask here mostly for safety:
        x = torch.where(mask.unsqueeze(-1), x, 0)
        return x


class ToVector(nn.Module):
    """Per-unit processing then average over the units dimension."""

    def __init__(
        self,
        input_dim: int,
        hidden_dim: int,
        output_dim: int,
        use_layer_norm: bool = True,
    ):
        super().__init__()

        self.mlp = MLP([input_dim, hidden_dim])
        self.out = nn.Sequential(
            *(
                ([nn.LayerNorm(hidden_dim)] if use_layer_norm else [])
                + [nn.ReLU(), nn.Linear(hidden_dim, output_dim)]
            )
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.mlp(x)
        # x = torch.cat((x[..., :1, :].mean(1), x[..., 1:, :].mean(1)), dim=-1)
        x = x.mean(1)
        return self.out(x)


class VectorMerge(nn.Module):
    def __init__(
        self,
        input_sizes: Mapping[str, Optional[int]],
        output_size: int,
        gating_type: str = "none",
        use_layer_norm: bool = True,
        name: Optional[str] = None,
    ):
        super().__init__()

        if not input_sizes:
            raise ValueError("input_names cannot be empty")
        self._input_sizes = input_sizes
        self._output_size = output_size
        self._gating_type = gating_type
        self._use_layer_norm = use_layer_norm
        self._input_dtypes = {}

        self.pre_layers = nn.ModuleDict(
            {
                key: nn.Sequential(
                    *(([nn.LayerNorm(value)] if use_layer_norm else []) + [nn.ReLU()])
                )
                for key, value in input_sizes.items()
            }
        )
        self.encoding_layers = nn.ModuleDict(
            {key: nn.Linear(value, output_size) for key, value in input_sizes.items()}
        )
        self.gating_layers = nn.ModuleDict(
            {
                key: nn.Linear(value, len(input_sizes) * value)
                for key, value in input_sizes.items()
            }
        )

    def _compute_gate(
        self,
        inputs_to_gate: Dict[str, torch.Tensor],
        init_gate: Dict[str, torch.Tensor],
    ):
        leading_dims = next(iter(inputs_to_gate.values())).shape[:2]
        gate = {k: self.gating_layers[k](y) for k, y in init_gate.items()}
        gate = sum(gate.values())
        gate = gate.view(*leading_dims, len(inputs_to_gate), self._output_size)
        gate = torch.softmax(gate, dim=len(leading_dims))
        gate = {k: gate[:, :, i] for i, (k, _) in enumerate(inputs_to_gate.items())}
        return gate

    def _encode(self, inputs: Dict[str, torch.Tensor]):
        gate, outputs = {}, {}
        for name, _ in self._input_sizes.items():
            feature = inputs[name]
            feature = self.pre_layers[name](feature)
            gate[name] = feature
            outputs[name] = self.encoding_layers[name](feature)
        return gate, outputs

    def forward(self, inputs: Dict[str, torch.Tensor]):
        gate, outputs = self._encode(inputs)
        if len(outputs) == 1:
            # Special case of 1-D inputs that do not need any gating.
            output = outputs[0]
        else:
            gate = self._compute_gate(outputs, gate)
            data = [gate[k] * outputs[k] for k in inputs.keys()]
            output = sum(data)
        return output



import os

os.environ["OMP_NUM_THREADS"] = "1"

import wandb
import threading
import multiprocessing as mp

from typing import List

from tqdm import tqdm

from meloetta.workers import SelfPlayWorker, EvalWorker

from meloetta.frameworks.random import RandomActor
from meloetta.frameworks.max_damage import MaxDamageActor
from meloetta.frameworks.nash_ketchum import NAshKetchumActor, NAshKetchumLearner


def start_frame_monitor(queue: mp.Queue):
    frame_monitor = tqdm(desc="frames: ")
    episode_monitor = tqdm(desc="eps: ")
    while True:
        _, frames = queue.get()
        frame_monitor.update(frames)
        episode_monitor.update(1)
        if not frames:
            break


def start_eval_monitor(queue: mp.Queue):
    random = 0
    maxdmg = 0
    while True:
        baseline, res = queue.get()
        if baseline == "random":
            n = random
            random += 1
        elif baseline == "maxdmg":
            n = maxdmg
            maxdmg += 1
        wandb.log({f"{baseline}n": n, baseline: 1 - res})


def main(fpath: str = None, latest: bool = False):
    if latest:
        print("Using latest save checkpoint")
        learner = NAshKetchumLearner.from_latest()

    elif fpath:
        print(f"Loading run from: {fpath}")
        learner = NAshKetchumLearner.from_fpath(fpath)

    else:
        learner = NAshKetchumLearner()

    wandb.init(
        project="meloetta",
        config=learner.get_config(),
    )

    eval_queue = mp.Queue()

    main_actor = NAshKetchumActor
    random_actor = RandomActor
    maxdmg_actor = MaxDamageActor

    procs: List[mp.Process] = []

    if learner.config.eval:
        evals = [
            (f"random_eval{i}", f"random{i}", random_actor, (eval_queue,))
            for i in range(0)
        ] + [
            (
                f"maxdmg_eval{i}",
                f"maxdmg{i}",
                maxdmg_actor,
                (learner.config.gen, eval_queue),
            )
            for i in range(1)
        ]
        for i, (
            eval_username,
            opponent_username,
            opponent_actor,
            opponent_actor_args,
        ) in enumerate(evals):
            worker = EvalWorker(
                eval_username=eval_username,
                opponent_username=opponent_username,
                battle_format=learner.config.battle_format,
                team=learner.config.team,
                eval_actor_fn=main_actor,
                eval_actor_kwargs={
                    "model": learner.actor_model,
                    # "replay_buffer": learner.replay_buffer,
                },
                baseline_actor_fn=opponent_actor,
                baseline_actor_args=opponent_actor_args,
            )
            process = mp.Process(
                target=worker.run,
                name=repr(worker) + str(i),
            )
            process.start()
            procs.append(process)

    threads: List[threading.Thread] = []

    frame_monitor_thread = threading.Thread(
        target=start_frame_monitor, args=(learner.replay_buffer.finish_queue,)
    )
    frame_monitor_thread.start()
    threads.append(frame_monitor_thread)

    if not learner.config.eval_mode:
        for i in range(1):
            learner_thread = threading.Thread(
                target=learner.run,
                name=f"Learning Thread{i}",
            )
            threads.append(learner_thread)
            learner_thread.start()

        if learner.config.debug_mode:
            num_players = 2
        else:
            num_players = max(learner.config.batch_size // learner.config.num_actors, 2)

        for i in range(learner.config.num_actors):
            worker = SelfPlayWorker(
                worker_index=i,
                num_players=num_players,  # 2 is players per worker
                battle_format=learner.config.battle_format,
                team=learner.config.team,
                actor_fn=main_actor,
                actor_kwargs={
                    "model": learner.actor_model,
                    "replay_buffer": learner.replay_buffer,
                },
            )

            process = mp.Process(
                target=worker.run,
                name=repr(worker) + str(i),
            )
            process.start()
            procs.append(process)

    eval_monitor_thread = threading.Thread(
        target=start_eval_monitor, args=(eval_queue,)
    )
    eval_monitor_thread.start()
    threads.append(eval_monitor_thread)

    for thread in threads:
        thread.join()

    for proc in procs:
        proc.join()


if __name__ == "__main__":
    mp.set_start_method("spawn")

    # fpath = "cpkts/cpkt-101000.tar"
    # main(fpath)

    # main(latest=True)

    main()



from typing import NamedTuple


class _Config(NamedTuple):
    @classmethod
    def from_dict(self, d: dict):
        return self(**d)


class PrivateEncoderConfig(_Config):
    embedding_dim: int = 128
    entity_embedding_dim: int = 128
    transformer_num_heads: int = 2
    transformer_num_layers: int = 3
    resblocks_num_before: int = 1
    resblocks_num_after: int = 1
    output_dim: int = embedding_dim * 2


class PublicEncoderConfig(_Config):
    scalar_embedding_dim: int = 128
    entity_embedding_dim: int = 128
    transformer_num_heads: int = 2
    transformer_num_layers: int = 3
    resblocks_num_before: int = 1
    resblocks_num_after: int = 1
    output_dim: int = entity_embedding_dim * 2


class ScalarEncoderConfig(_Config):
    embedding_dim: int = 32


class WeatherEncoderConfig(_Config):
    embedding_dim: int = 32


class EncoderConfig(_Config):
    private_encoder_config: PrivateEncoderConfig = PrivateEncoderConfig()
    public_encoder_config: PublicEncoderConfig = PublicEncoderConfig()
    scalar_encoder_config: ScalarEncoderConfig = ScalarEncoderConfig()
    weather_encoder_config: WeatherEncoderConfig = WeatherEncoderConfig()
    output_dim: int = (
        private_encoder_config.output_dim  # private encoder
        + private_encoder_config.entity_embedding_dim  # private spatial
        + public_encoder_config.output_dim  # public encoder
        + public_encoder_config.output_dim  # public spatial
        + weather_encoder_config.embedding_dim
        + scalar_encoder_config.embedding_dim
    )


class CoreConfig(_Config):
    raw_embedding_dim: int = EncoderConfig.output_dim
    hidden_dim: int = 384
    num_layers: int = 3


class ValueHeadConfig(_Config):
    state_embedding_dim: int = CoreConfig.hidden_dim
    hidden_dim: int = 256
    num_resblocks: int = 2


class ActionTypeHeadConfig(_Config):
    state_embedding_dim: int = CoreConfig.hidden_dim
    num_action_types: int = 3
    context_dim: int = ScalarEncoderConfig.embedding_dim
    residual_dim: int = 256
    action_map_dim: int = 256
    autoregressive_embedding_dim: int = 256


class FlagHeadConfig(_Config):
    autoregressive_embedding_dim: int = (
        ActionTypeHeadConfig.autoregressive_embedding_dim
    )
    hidden_dim: int = 256


class MaxMoveHeadConfig(_Config):
    entity_embedding_dim: int = PrivateEncoderConfig.entity_embedding_dim
    key_dim: int = 32

    autoregressive_embedding_dim: int = (
        ActionTypeHeadConfig.autoregressive_embedding_dim
    )
    query_hidden_dim: int = 256


class MoveHeadConfig(_Config):
    entity_embedding_dim: int = PrivateEncoderConfig.entity_embedding_dim
    key_dim: int = 32

    autoregressive_embedding_dim: int = (
        ActionTypeHeadConfig.autoregressive_embedding_dim
    )
    query_hidden_dim: int = 256


class SwitchHeadConfig(_Config):
    entity_embedding_dim: int = PrivateEncoderConfig.entity_embedding_dim
    key_dim: int = 32

    autoregressive_embedding_dim: int = (
        ActionTypeHeadConfig.autoregressive_embedding_dim
    )
    query_hidden_dim: int = 256


class TargetHeadConfig(_Config):
    pass


class PolicyHeadsConfig(_Config):
    action_type_head_config: ActionTypeHeadConfig = ActionTypeHeadConfig()
    flag_head_config: FlagHeadConfig = FlagHeadConfig()
    max_move_head_config: MaxMoveHeadConfig = MaxMoveHeadConfig()
    move_head_config: MoveHeadConfig = MoveHeadConfig()
    switch_head_config: SwitchHeadConfig = SwitchHeadConfig()
    target_head_config: TargetHeadConfig = TargetHeadConfig()


def _scale(config: _Config, factor: float):
    datum = {}
    for key, value in config._asdict().items():
        if isinstance(value, tuple) and hasattr(value, "_asdict"):
            datum[key] = _scale(value, factor)
        elif isinstance(value, int) or isinstance(value, float):
            datum[key] = max(int(value * factor), 1)

    return config.from_dict(datum)


class MewZeroModelConfig(_Config):
    encoder_config: EncoderConfig = EncoderConfig()
    policy_heads_config: PolicyHeadsConfig = PolicyHeadsConfig()
    resnet_core_config: CoreConfig = CoreConfig()
    value_head_config: ValueHeadConfig = ValueHeadConfig()

    @classmethod
    def from_scale(self, scale: float):
        return _scale(self(), scale)


def main():
    config = MewZeroModelConfig.from_scale(0.2)
    print(config)


if __name__ == "__main__":
    main()



import torch
import torch.nn as nn

from meloetta.frameworks.nash_ketchum.model import config
from meloetta.frameworks.nash_ketchum.model.utils import MLP
from meloetta.data import WEATHERS, PSEUDOWEATHERS


class WeatherEncoder(nn.Module):
    def __init__(self, gen: int, config: config.WeatherEncoderConfig):
        super().__init__()

        self.weather_onehot = nn.Embedding.from_pretrained(torch.eye(len(WEATHERS) + 1))
        self.time_left_onehot = nn.Embedding.from_pretrained(torch.eye(10))
        self.min_time_left_onehot = nn.Embedding.from_pretrained(torch.eye(7))

        pw_min_onehot = nn.Embedding.from_pretrained(torch.eye(8))
        self.pw_min_onehot = nn.Sequential(pw_min_onehot, nn.Flatten(2))
        pw_max_onehot = nn.Embedding.from_pretrained(torch.eye(10))
        self.pw_max_onehot = nn.Sequential(pw_max_onehot, nn.Flatten(2))

        lin_in = (
            self.weather_onehot.embedding_dim
            + self.time_left_onehot.embedding_dim
            + self.min_time_left_onehot.embedding_dim
            + pw_min_onehot.embedding_dim * len(PSEUDOWEATHERS)
            + pw_max_onehot.embedding_dim * len(PSEUDOWEATHERS)
        )
        self.encoder = MLP([lin_in, config.embedding_dim, config.embedding_dim])

    def forward(
        self,
        weather: torch.Tensor,
        pseudoweather: torch.Tensor,
    ):
        weather_token = weather[..., 0]
        time_left = weather[..., 1]
        min_time_left = weather[..., 2]

        weather_onehot = self.weather_onehot(weather_token + 1)
        time_left_onehot = self.time_left_onehot(time_left)
        min_time_left_onehot = self.min_time_left_onehot(min_time_left)

        pseudo_weather_x = pseudoweather + 1
        pw_min_time_left = pseudo_weather_x[..., 0]
        pw_max_time_left = pseudo_weather_x[..., 1]

        pw_min_time_left_onehot = self.pw_min_onehot(pw_min_time_left)
        pw_max_time_left_onehot = self.pw_max_onehot(pw_max_time_left)

        weather_raw = torch.cat(
            (
                weather_onehot,
                time_left_onehot,
                min_time_left_onehot,
                pw_max_time_left_onehot,
                pw_min_time_left_onehot,
            ),
            dim=-1,
        )
        return self.encoder(weather_raw)



from meloetta.frameworks.nash_ketchum.model.encoder import Encoder
from meloetta.frameworks.nash_ketchum.model.policy import PolicyHeads
from meloetta.frameworks.nash_ketchum.model.core import Core
from meloetta.frameworks.nash_ketchum.model.value import ValueHead
from meloetta.frameworks.nash_ketchum.model.heads.move_head import MoveHead
from meloetta.frameworks.nash_ketchum.model.heads.switch_head import SwitchHead
from meloetta.frameworks.nash_ketchum.model.heads.flag_head import FlagsHead
from meloetta.frameworks.nash_ketchum.model.interfaces import *
from meloetta.frameworks.nash_ketchum.model.utils import *
from meloetta.frameworks.nash_ketchum.model.main import NAshKetchumModel
from meloetta.frameworks.nash_ketchum.model.config import *



from setuptools import find_packages, setup

from meloetta import __version__

requirements = [
    "beautifulsoup4==4.12.2",
    "jax==0.3.14",
    "detoxify==0.5.1",
    "pandas==1.3.5",
    "numpy==1.21.5",
    "py_mini_racer==0.6.0",
    "plotly==5.11.0",
    "scikit_learn==1.2.2",
    "requests==2.26.0",
    "torch==1.13.0",
    "setuptools==58.0.0",
    "tqdm==4.64.1",
    "websockets==10.1",
]


pretrain_requirements = [
    "transformers[torch]==4.18.0",
    "sentence-transformers==2.2.0",
    "huggingface-hub==0.4.0",
]


long_description = """Meloetta is a Pokémon Battle Client for interacting with Pokémon Showdown written in Python. This project was born out of frustration for currently existing tools and their lack of dependency on Zarel's (PS Creator) existing code for handling client server interation.

The client works by reading messages from an asyncio stream and forwarding these to the javascript client battle object with PyMiniRacer. This concept was taken from [metagrok](https://github.com/yuzeh/metagrok).

As is, the necessary javascript source files come with the pip install. Whenever Pokemon Showdown Client releases an update, the client code can also be automatically updated from the source."""

setup(
    name="meloetta",
    version=__version__,
    url="https://github.com/spktrm/meloetta",
    author="Joseph Twin",
    author_email="joseph.twin14@gmail.com",
    long_description=long_description,
    long_description_content_type="text/markdown",
    include_package_data=True,
    packages=find_packages(),
    install_requires=requirements,
    extras_require=dict(
        core_tests=pretrain_requirements,
    ),
)



import torch
import torch.nn as nn
import torch.nn.functional as F

from typing import Tuple, List

from meloetta.frameworks.porygon.model.interfaces import EncoderOutput

from meloetta.frameworks.porygon.model import config
from meloetta.frameworks.porygon.model.utils import Resblock


def script_lnlstm(
    input_size,
    hidden_size,
    num_layers,
    bias=True,
    batch_first=False,
    dropout=False,
) -> "StackedLSTM":
    """Returns a ScriptModule that mimics a PyTorch native LSTM."""

    # The following are not implemented.
    assert bias
    assert not batch_first
    assert not dropout

    stack_type = StackedLSTM
    layer_type = LSTMLayer
    dirs = 1

    return stack_type(
        num_layers,
        layer_type,
        first_layer_args=[
            LayerNormLSTMCell,
            input_size,
            hidden_size,
        ],
        other_layer_args=[
            LayerNormLSTMCell,
            hidden_size * dirs,
            hidden_size,
        ],
    )


def init_stacked_lstm(num_layers, layer, first_layer_args, other_layer_args):
    layers = [layer(*first_layer_args)] + [
        layer(*other_layer_args) for _ in range(num_layers - 1)
    ]
    return nn.ModuleList(layers)


class LayerNormLSTMCell(nn.Module):
    def __init__(self, input_size, hidden_size):
        super(LayerNormLSTMCell, self).__init__()
        self.input_size = input_size
        self.hidden_size = hidden_size
        self.weight_ih = nn.Parameter(torch.randn(4 * hidden_size, input_size))
        self.weight_hh = nn.Parameter(torch.randn(4 * hidden_size, hidden_size))
        # The layernorms provide learnable biases
        ln = nn.LayerNorm

        self.layernorm_i = ln(4 * hidden_size)
        self.layernorm_h = ln(4 * hidden_size)
        self.layernorm_c = ln(hidden_size)

    def forward(
        self, input: torch.Tensor, state: Tuple[torch.Tensor, torch.Tensor]
    ) -> Tuple[torch.Tensor, Tuple[torch.Tensor, torch.Tensor]]:
        hx, cx = state
        igates = self.layernorm_i(torch.mm(input, self.weight_ih.t()))
        hgates = self.layernorm_h(torch.mm(hx, self.weight_hh.t()))
        gates = igates + hgates
        ingate, forgetgate, cellgate, outgate = gates.chunk(4, 1)

        ingate = torch.sigmoid(ingate)
        forgetgate = torch.sigmoid(forgetgate)
        cellgate = torch.tanh(cellgate)
        outgate = torch.sigmoid(outgate)

        cy = self.layernorm_c((forgetgate * cx) + (ingate * cellgate))
        hy = outgate * torch.tanh(cy)

        return hy, (hy, cy)


class LSTMLayer(nn.Module):
    def __init__(self, cell, *cell_args):
        super(LSTMLayer, self).__init__()
        self.cell = cell(*cell_args)

    def forward(
        self, input: torch.Tensor, state: Tuple[torch.Tensor, torch.Tensor]
    ) -> Tuple[torch.Tensor, Tuple[torch.Tensor, torch.Tensor]]:
        inputs = input.unbind(0)
        outputs = torch.jit.annotate(List[torch.Tensor], [])
        for i in range(len(inputs)):
            out, state = self.cell(inputs[i], state)
            outputs += [out]
        return torch.stack(outputs), state


class StackedLSTM(nn.Module):
    __constants__ = ["layers"]  # Necessary for iterating through self.layers

    def __init__(self, num_layers, layer, first_layer_args, other_layer_args):
        super(StackedLSTM, self).__init__()
        self.layers = init_stacked_lstm(
            num_layers, layer, first_layer_args, other_layer_args
        )

    def forward(
        self, input: torch.Tensor, states: List[Tuple[torch.Tensor, torch.Tensor]]
    ) -> Tuple[torch.Tensor, List[Tuple[torch.Tensor, torch.Tensor]]]:
        # List[LSTMState]: One state per layer
        h_out = []
        c_out = []
        output = input
        # XXX: enumerate https://github.com/pytorch/pytorch/issues/14471
        i = 0
        for rnn_layer, hi, ci in zip(self.layers, *states):
            state = (hi, ci)
            output, (ho, co) = rnn_layer(output, state)
            h_out += [ho]
            c_out += [co]
            i += 1
        h_out = torch.stack(h_out)
        c_out = torch.stack(c_out)
        return output, (h_out, c_out)


class CoreResblock(nn.Module):
    def __init__(self, in_features, out_features):
        super().__init__()

        self.lin1 = nn.Sequential(
            nn.Linear(in_features, out_features),
            nn.LayerNorm(out_features),
            nn.ReLU(),
        )
        self.lin2 = nn.Sequential(
            nn.Linear(out_features, out_features),
            nn.LayerNorm(out_features),
        )
        if out_features != in_features:
            self.res = nn.Linear(in_features, out_features)
        else:
            self.res = None

    def forward(self, x):
        if self.res:
            res = self.res(x)
        else:
            res = x
        x = self.lin1(x)
        x = self.lin2(x)
        return F.relu(x) + res


class Core(nn.Module):
    def __init__(self, config: config.CoreConfig):
        super().__init__()
        self.config = config

        self.mlp1 = CoreResblock(config.raw_embedding_dim, config.hidden_dim)
        # self.mlp2 = CoreResblock(config.hidden_dim, config.hidden_dim)
        # self.mlp3 = CoreResblock(config.hidden_dim, config.hidden_dim)

        # self.rnn = script_lnlstm(
        #     config.raw_embedding_dim,
        #     config.hidden_dim,
        #     num_layers=config.num_layers,
        # )

        # self.rnn = nn.GRU(
        #     input_size=config.raw_embedding_dim,
        #     hidden_size=config.hidden_dim,
        #     num_layers=config.num_layers,
        # )

    def initial_state(self, batch_size: int) -> Tuple[torch.Tensor, torch.Tensor]:
        # def initial_state(self, batch_size: int) -> torch.Tensor:
        return tuple(
            torch.zeros(self.config.num_layers, batch_size, self.config.hidden_dim)
            for _ in range(2)
        )
        # return torch.zeros(self.config.num_layers, batch_size, self.config.hidden_dim)

    def forward(
        self,
        encoder_output: EncoderOutput,
        hidden_state: Tuple[torch.Tensor, torch.Tensor],
    ):
        state_embedding = torch.cat(
            (
                encoder_output.side_embs.side_embedding.flatten(2),
                encoder_output.side_embs.side_context,
                encoder_output.weather_emb,
                encoder_output.scalar_emb,
            ),
            dim=-1,
        )

        state_embedding = self.mlp1(state_embedding)
        # state_embedding = self.mlp2(state_embedding)
        # state_embedding = self.mlp3(state_embedding)

        # state_embedding, hidden_state = self.rnn(state_embedding, hidden_state)

        return state_embedding, hidden_state



import math
import json
import torch
import random
import traceback

import torch.nn as nn

from copy import deepcopy
from typing import Optional, Union, Literal, Tuple, Dict, Any, List

from meloetta.actors.types import State, Choices

from meloetta.frameworks.porygon.model.utils import _log_policy

from meloetta.frameworks.porygon.model import config
from meloetta.frameworks.porygon.model import (
    Encoder,
    PolicyHeads,
    Core,
    ValueHead,
    Indices,
    Logits,
    Policy,
    LogPolicy,
    TrainingOutput,
    EncoderOutput,
    ModelOutput,
    PostProcess,
    Batch,
)

from meloetta.data import _STATE_FIELDS


class PorygonModel(nn.Module):
    def __init__(
        self,
        gen: int = 9,
        gametype: Literal["singles", "doubles", "triples"] = "singles",
        config: config.PorygonModelConfig = config.PorygonModelConfig(),
    ) -> None:
        super().__init__()

        self.gen = gen
        self.gametype = gametype
        n_active = {"singles": 1, "doubles": 2, "triples": 3}[gametype]

        self.encoder = Encoder(gen=gen, n_active=n_active, config=config.encoder_config)
        self.policy_heads = PolicyHeads(
            gen=gen, gametype=gametype, config=config.policy_heads_config
        )

        self.state_fields = deepcopy(_STATE_FIELDS)

        if gen != 8:
            self.state_fields.remove("max_move_mask")

        if self.gametype == "singles":
            self.state_fields.remove("target_mask")
            self.state_fields.remove("prev_choices")

        self.core = Core(config.resnet_core_config)
        self.value_Head = ValueHead(config.value_head_config)

    def get_learnable_params(self) -> int:
        return sum([p.numel() for p in self.parameters() if p.requires_grad])

    def clean(self, state: State) -> State:
        return {k: state[k] for k in self.state_fields}

    def forward(
        self,
        state: State,
        hidden_state: Tuple[torch.Tensor, torch.Tensor] = None,
        choices: Choices = None,
        need_log_policy: bool = True,
        battle_tag: str = None,
    ):
        encoder_output = self.encoder.forward(state)
        state_emb, hidden_state = self.core.forward(encoder_output, hidden_state)
        value = self.value_Head.forward(state_emb)
        while True:
            indices, logits, policy = self.policy_heads.forward(
                state_emb,
                encoder_output,
                state,
            )
            try:
                post_process = self._postprocess(
                    state=state,
                    encoder_output=encoder_output,
                    hidden_state=hidden_state,
                    indices=indices,
                    logits=logits,
                    policy=policy,
                    value=value,
                    choices=choices,
                    need_log_policy=need_log_policy,
                )
            except Exception:
                traceback.print_exc()

                with open(
                    f"{battle_tag}_model_error"
                    + str(random.randint(0, 9999999))
                    + ".json",
                    "w",
                ) as f:
                    json.dump(
                        {
                            "logits": {
                                k: torch.squeeze(v).tolist()
                                for k, v in logits._asdict().items()
                                if v is not None
                            },
                            "policy": {
                                k: torch.squeeze(v).tolist()
                                for k, v in policy._asdict().items()
                                if v is not None
                            },
                            "indices": {
                                k: torch.squeeze(v).tolist()
                                for k, v in indices._asdict().items()
                                if v is not None
                            },
                        },
                        f,
                    )
            else:
                return post_process

    def learning_forward(
        self,
        batch: Batch,
    ):
        T, B = batch.trajectory_length, batch.batch_size

        outputs: List[TrainingOutput] = []

        minibatch_size = min(128, (4096 // batch.trajectory_length) + 1)

        for batch_index in range(math.ceil(batch.batch_size / minibatch_size)):
            batch_start = minibatch_size * batch_index
            batch_end = minibatch_size * (batch_index + 1)

            minibatch = {
                key: value[:, batch_start:batch_end].contiguous()
                for key, value in batch._asdict().items()
                if value is not None
            }
            output, _, decoder = self.forward(minibatch, None)
            outputs.append(output)

        pi_dict: Dict[str, List[torch.Tensor]] = {}
        log_pi_dict: Dict[str, List[torch.Tensor]] = {}
        logit_dict: Dict[str, List[torch.Tensor]] = {}
        vs: List[torch.Tensor] = []

        for o in outputs:
            for key in Policy._fields:
                mbv = getattr(o.pi, key)
                if getattr(o.pi, key) is not None:
                    if key not in pi_dict:
                        pi_dict[key] = []
                    pi_dict[key].append(mbv)

            for key in LogPolicy._fields:
                mbv = getattr(o.log_pi, key)
                if getattr(o.log_pi, key) is not None:
                    if key not in log_pi_dict:
                        log_pi_dict[key] = []
                    log_pi_dict[key].append(mbv)

            for key in Logits._fields:
                mbv = getattr(o.logit, key)
                if getattr(o.logit, key) is not None:
                    if key not in logit_dict:
                        logit_dict[key] = []
                    logit_dict[key].append(mbv)

            vs.append(o.v)

        pi_dict = {
            key: torch.cat(value, dim=1)
            .contiguous()
            .view(T, B, *value[0].shape[2:])
            .squeeze()
            for key, value in pi_dict.items()
        }
        log_pi_dict = {
            key: torch.cat(value, dim=1)
            .contiguous()
            .view(T, B, *value[0].shape[2:])
            .squeeze()
            for key, value in log_pi_dict.items()
        }
        logit_dict = {
            key: torch.cat(value, dim=1)
            .contiguous()
            .view(T, B, *value[0].shape[2:])
            .squeeze()
            for key, value in logit_dict.items()
        }
        vs = torch.cat(vs, dim=1).contiguous().view(T, B, 1)

        return TrainingOutput(
            pi=Policy(**pi_dict),
            v=vs,
            logit=Logits(**logit_dict),
            log_pi=LogPolicy(**log_pi_dict),
        )

    def _postprocess(
        self,
        state: State,
        encoder_output: EncoderOutput,
        hidden_state: Tuple[torch.Tensor, torch.Tensor],
        indices: Indices,
        logits: Logits,
        policy: Policy,
        value: torch.Tensor,
        choices: Optional[Dict[str, Any]] = None,
        need_log_policy: bool = True,
    ) -> Union[
        Tuple[TrainingOutput, Tuple[torch.Tensor, torch.Tensor], EncoderOutput],
        Tuple[ModelOutput, PostProcess, Tuple[torch.Tensor, torch.Tensor]],
    ]:
        if not self.training:
            targeting = bool(state.get("targeting"))
            post_process = self.index_to_action(targeting, indices, choices)
            env_step = ModelOutput(
                indices=indices,
                policy=policy,
                value=value,
                logits=logits,
            )
            output = (env_step, post_process, hidden_state)
        else:
            if need_log_policy:
                log_policy = self.get_log_policy(logits, state)
            else:
                log_policy = None
            output = (
                TrainingOutput(
                    pi=policy,
                    v=value,
                    log_pi=log_policy,
                    logit=logits,
                ),
                hidden_state,
                encoder_output,
            )
        return output

    def get_log_policy(self, logits: Logits, state: State):
        if self.gametype != "singles":
            target_log_policy = _log_policy(
                logits.target_logits,
                state["target_mask"],
            )
        else:
            target_log_policy = None

        if self.gen == 8:
            max_move_log_policy = _log_policy(
                logits.max_move_logits,
                state["max_move_mask"],
            )
        else:
            max_move_log_policy = None

        if self.gen >= 6:
            flag_log_policy = _log_policy(
                logits.flag_logits,
                state["flag_mask"],
            )
        else:
            flag_log_policy = None

        return LogPolicy(
            action_type_log_policy=_log_policy(
                logits.action_type_logits, state["action_type_mask"]
            ),
            move_log_policy=_log_policy(logits.move_logits, state["move_mask"]),
            switch_log_policy=_log_policy(logits.switch_logits, state["switch_mask"]),
            max_move_log_policy=max_move_log_policy,
            flag_log_policy=flag_log_policy,
            target_log_policy=target_log_policy,
        )

    def index_to_action(
        self,
        targeting: bool,
        indices: Indices,
        choices: Optional[Dict[str, Any]] = None,
    ):
        action_type_index = indices.action_type_index.item()
        move_index = indices.move_index.item()
        switch_index = indices.switch_index.item()

        if not targeting:
            index = action_type_index

            if action_type_index == 0:
                flag_index = (indices.flag_index).item()
                if flag_index == 3:
                    max_move_index = (indices.max_move_index).item()
                    index = max_move_index
                else:
                    index = move_index

                if choices is not None:
                    if flag_index == 0:
                        data = choices["moves"]
                    elif flag_index == 1:
                        data = choices["mega_moves"]
                    elif flag_index == 2:
                        data = choices["zmoves"]
                    elif flag_index == 3:
                        data = choices["max_moves"]
                    elif flag_index == 4:
                        data = choices["tera_moves"]
                    else:
                        data = choices["moves"]
                else:
                    data = None

            elif action_type_index == 1:
                if choices is not None:
                    data = choices["switches"]
                else:
                    data = None
                index = switch_index

        else:
            if choices is not None:
                data = choices["targets"]
            else:
                data = None

            target_index = (indices.target_index).item()
            index = target_index

        return PostProcess(data, index)



from meloetta.frameworks.mewzero.model.encoders.private_encoder import *
from meloetta.frameworks.mewzero.model.encoders.public_encoder import *
from meloetta.frameworks.mewzero.model.encoders.weather_encoder import (
    WeatherEncoder,
)
from meloetta.frameworks.mewzero.model.encoders.scalar_encoder import ScalarEncoder



import torch

from torch import nn
from torch.nn import functional as F

from typing import Tuple

from meloetta.frameworks.nash_ketchum.model import config
from meloetta.frameworks.nash_ketchum.model.utils import (
    sqrt_one_hot_matrix,
    power_one_hot_matrix,
    binary_enc_matrix,
    TransformerEncoder,
    MLP,
    ToVector,
    VectorMerge,
    GLU,
)
from meloetta.frameworks.nash_ketchum.model.interfaces import SideEncoderOutput
from meloetta.embeddings import (
    AbilityEmbedding,
    PokedexEmbedding,
    MoveEmbedding,
    ItemEmbedding,
)

from meloetta.data import (
    GENDERS,
    BOOSTS,
    VOLATILES,
    STATUS,
    TOKENIZED_SCHEMA,
    ITEM_EFFECTS,
    SIDE_CONDITIONS,
    BattleTypeChart,
)


class PokemonEmbedding(nn.Module):
    def __init__(self, gen: int = 9, output_dim: int = 128) -> None:
        super().__init__()

        self.gen = gen

        pokedex_embedding = PokedexEmbedding(gen=gen)
        self.pokedex_embedding = nn.Sequential(
            pokedex_embedding, nn.Linear(pokedex_embedding.embedding_dim, output_dim)
        )
        ability_embedding = AbilityEmbedding(gen=gen)
        self.ability_embedding = nn.Sequential(
            ability_embedding, nn.Linear(ability_embedding.embedding_dim, output_dim)
        )
        self.item_embedding = ItemEmbedding(gen=gen)
        self.item_effect_onehot = nn.Embedding.from_pretrained(
            torch.eye(len(ITEM_EFFECTS) + 1)[..., 1:]
        )
        self.item_lin = nn.Linear(
            self.item_embedding.embedding_dim + self.item_effect_onehot.embedding_dim,
            output_dim,
        )
        self.pp_bin_enc = nn.Embedding.from_pretrained(binary_enc_matrix(64))
        self.move_embedding_ae = MoveEmbedding(gen=gen)
        self.move_embedding = nn.Sequential(
            nn.Linear(
                self.move_embedding_ae.embedding_dim + self.pp_bin_enc.embedding_dim,
                output_dim,
            )
        )
        self.last_move_embedding = nn.Linear(
            self.move_embedding_ae.embedding_dim + self.pp_bin_enc.embedding_dim,
            output_dim,
        )

        self.active_onehot = nn.Embedding.from_pretrained(torch.eye(3)[..., 1:])
        self.fainted_onehot = nn.Embedding.from_pretrained(torch.eye(3)[..., 1:])
        self.gender_onehot = nn.Embedding.from_pretrained(
            torch.eye(len(GENDERS) + 1)[..., 1:]
        )

        status_dummy = torch.eye(len(STATUS) + 1)[..., 1:]
        self.status_onehot = nn.Embedding.from_pretrained(status_dummy)
        self.sleep_turns_onehot = nn.Embedding.from_pretrained(torch.eye(4)[..., 1:])
        self.toxic_turns_onehot = nn.Embedding.from_pretrained(
            sqrt_one_hot_matrix(16)[..., 1:]
        )

        num_formes = len(TOKENIZED_SCHEMA[f"gen{gen}"]["pokedex"]["forme"]) + 1
        self.forme_embedding = nn.Embedding.from_pretrained(
            torch.eye(num_formes)[..., 1:]
        )
        self.level_onehot = nn.Embedding.from_pretrained(torch.eye(100))

        self.hp_sqrt_onehot = nn.Embedding.from_pretrained(
            sqrt_one_hot_matrix(768 if gen != 8 else 1536)[..., 1:]
        )
        self.stat_sqrt_onehot = nn.Embedding.from_pretrained(
            power_one_hot_matrix(512, 1 / 3)[..., 1:]
        )

        self.side_onehot = nn.Embedding.from_pretrained(torch.eye(2))
        self.known_onehot = nn.Embedding.from_pretrained(torch.eye(2))

        onehot_size = (
            self.active_onehot.embedding_dim
            + self.fainted_onehot.embedding_dim
            + self.gender_onehot.embedding_dim
            + self.status_onehot.embedding_dim
            + self.sleep_turns_onehot.embedding_dim
            + self.toxic_turns_onehot.embedding_dim
            + self.forme_embedding.embedding_dim
            + self.level_onehot.embedding_dim
            + 2 * self.hp_sqrt_onehot.embedding_dim
            + 1
            + 5 * self.stat_sqrt_onehot.embedding_dim
            + self.side_onehot.embedding_dim
            + self.known_onehot.embedding_dim
        )

        if gen == 9:
            self.commanding_onehot = nn.Embedding.from_pretrained(torch.eye(3)[..., 1:])
            self.reviving_onehot = nn.Embedding.from_pretrained(torch.eye(3)[..., 1:])
            self.tera_onehot = nn.Embedding.from_pretrained(torch.eye(2))
            self.teratype_onehot = nn.Embedding.from_pretrained(
                torch.eye(len(BattleTypeChart) + 1)[..., 1:]
            )

            onehot_size += (
                self.tera_onehot.embedding_dim + self.teratype_onehot.embedding_dim
            )

        elif gen == 8:
            self.can_gmax_embedding = nn.Embedding.from_pretrained(torch.eye(3))

            onehot_size += self.can_gmax_embedding.embedding_dim

        self.onehots_lin = nn.Linear(onehot_size, output_dim)

    def embed_moves(
        self, move_tokens: torch.Tensor, move_used: torch.Tensor
    ) -> torch.Tensor:
        moves_emb = self.move_embedding_ae(move_tokens)
        move_used_emb = self.pp_bin_enc(move_used)
        moves_emb = torch.cat((moves_emb, move_used_emb), dim=-1)
        return self.move_embedding(moves_emb)

    def embed_item(
        self, item_token: torch.Tensor, item_effect_token: torch.Tensor
    ) -> torch.Tensor:
        item_cat = torch.cat(
            (
                self.item_embedding(item_token),
                self.item_effect_onehot(item_effect_token),
            ),
            dim=-1,
        )
        return self.item_lin(item_cat)

    def forward(
        self, x: torch.Tensor
    ) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor,]:
        longs = (x + 1).long()

        name = longs[..., 0]
        forme = longs[..., 1]
        # slot = longs[..., 2]
        hp = longs[..., 3]
        maxhp = longs[..., 4]
        hp_ratio = x[..., 5]
        stats = longs[..., 6:11]
        fainted = longs[..., 11]
        active = longs[..., 12]
        level = x[..., 13].long()
        gender = longs[..., 14]
        ability = longs[..., 15]
        # base_ability = longs[..., 16]
        item = longs[..., 17]
        # prev_item = longs[..., 18]
        item_effect = longs[..., 19]
        # prev_item_effect = longs[..., 20]
        status = longs[..., 21]
        sleep_turns = longs[..., 22]
        toxic_turns = longs[..., 23]

        last_move = longs[..., 24]
        last_move_pp = longs[..., 25].clamp(max=63)

        moves = longs[..., 26:30]
        pp = longs[..., 30:34].clamp(max=63)

        if self.gen == 9:
            terastallized = longs[..., 33]
            teratype = longs[..., 35]
            # times_attacked = longs[..., 31]

        name_emb = self.pokedex_embedding(name)

        hp_emb = self.hp_sqrt_onehot(hp)
        maxhp_emb = self.hp_sqrt_onehot(maxhp)
        hp_ratio = hp_ratio.unsqueeze(-1)
        stat_onehot = self.stat_sqrt_onehot(stats).flatten(-2)

        ability_emb = self.ability_embedding(ability)
        # base_ability_emb = self.ability_embedding(base_ability)

        item_emb = self.embed_item(item, item_effect)
        # prev_item_emb = self.embed_item(prev_item, prev_item_effect)

        status_onehot = self.status_onehot(status)
        sleep_turns_onehot = self.sleep_turns_onehot(sleep_turns)
        toxic_turns_onehot = self.toxic_turns_onehot(toxic_turns)

        moves_emb = self.move_embedding(
            torch.cat((self.move_embedding_ae(moves), self.pp_bin_enc(pp)), dim=-1)
        )
        moveset_emb = moves_emb.sum(-2)

        last_move_emb = self.last_move_embedding(
            torch.cat(
                (self.move_embedding_ae(last_move), self.pp_bin_enc(last_move_pp)),
                dim=-1,
            )
        )

        side = torch.ones_like(active)
        side[:, :, :2] = 0

        known = torch.zeros_like(active)
        known[:, :, 1:] = 0

        forme_enc = self.forme_embedding(forme)
        stat_enc = torch.cat((hp_emb, maxhp_emb, hp_ratio, stat_onehot), dim=-1)
        active_enc = self.active_onehot(active)
        fainted_enc = self.fainted_onehot(fainted)
        gender_enc = self.gender_onehot(gender)
        level_enc = self.level_onehot(level.clamp(min=1) - 1)
        status_enc = torch.cat(
            (status_onehot, sleep_turns_onehot, toxic_turns_onehot), dim=-1
        )
        side_enc = self.side_onehot(side)
        known_enc = self.known_onehot(known)

        onehots = [
            forme_enc,
            stat_enc,
            active_enc,
            fainted_enc,
            gender_enc,
            level_enc,
            status_enc,
            side_enc,
            known_enc,
        ]

        if self.gen == 9:
            onehots += [self.teratype_onehot(teratype)]
            onehots += [self.tera_onehot((terastallized > 0).long())]

        onehots = torch.cat(onehots, dim=-1)
        onehots_emb = self.onehots_lin(onehots)

        pokemon_dict_embs = {
            "name_emb": name_emb,
            "ability_emb": ability_emb,
            "item_emb": item_emb,
            "moveset_emb": moveset_emb,
            "onehots": onehots_emb,
            # "last_move_emb": last_move_emb,
            # "base_ability_emb": base_ability_emb,
            # "prev_item_emb": prev_item_emb,
        }

        pokemon_emb = sum(pokemon_dict_embs.values())

        mask = name == 0  # | (fainted == 2)
        mask = ~mask

        pokemon_emb = pokemon_emb * mask.unsqueeze(-1)
        pokemon_emb = pokemon_emb.flatten(2, 3)

        mask = mask.flatten(2)

        return pokemon_emb, mask, moves_emb


class SideEncoder(nn.Module):
    def __init__(self, gen: int, n_active: int, config: config.SideEncoderConfig):
        super().__init__()

        self.config = config
        self.gen = gen
        self.n_active = n_active

        self.embedding = PokemonEmbedding(
            gen=gen, output_dim=config.entity_embedding_dim  # , frozen=True
        )

        self.transformer = TransformerEncoder(
            model_size=config.model_size,
            num_layers=config.num_layers,
            num_heads=config.num_heads,
            key_size=config.key_size,
            value_size=config.value_size,
            resblocks_num_before=config.resblocks_num_before,
            resblocks_num_after=config.resblocks_num_after,
            resblocks_hidden_size=config.resblocks_hidden_size,
            use_layer_norm=config.use_layer_norm,
        )

        self.output = ToVector(
            input_dim=config.entity_embedding_dim,
            hidden_dim=2 * config.entity_embedding_dim,
            output_dim=config.output_dim,
        )

        self.boosts1_onehot = nn.Embedding.from_pretrained(
            torch.tensor(
                [2 / n for n in range(8, 1, -1)] + [n / 2 for n in range(3, 9)]
            ).view(-1, 1)
        )
        self.boosts2_onehot = nn.Embedding.from_pretrained(
            torch.tensor(
                [3 / n for n in range(9, 2, -1)] + [n / 3 for n in range(4, 10)]
            ).view(-1, 1)
        )
        self.boosts_mlp = MLP([2 * len(BOOSTS), config.output_dim, config.output_dim])

        self.volatiles_mlp = MLP(
            [2 * len(VOLATILES), config.output_dim, config.output_dim]
        )

        self.toxicspikes_onehot = nn.Embedding.from_pretrained(torch.eye(3)[..., 1:])
        self.spikes_onehot = nn.Embedding.from_pretrained(torch.eye(4)[..., 1:])

        self.min_dur = nn.Embedding.from_pretrained(torch.eye(6)[..., 1:])
        self.max_dur = nn.Embedding.from_pretrained(torch.eye(9)[..., 1:])
        side_con_size = (
            2
            + self.toxicspikes_onehot.embedding_dim
            + self.spikes_onehot.embedding_dim
            + (len(SIDE_CONDITIONS) - 4)
            * (self.min_dur.weight.shape[-1] + self.max_dur.weight.shape[-1])
        )
        self.side_condt_mlp = MLP(
            [2 * side_con_size, config.output_dim, config.output_dim]
        )

        self.active_moves = GLU(
            config.entity_embedding_dim,
            config.entity_embedding_dim,
            config.entity_embedding_dim,
        )

    def encode_boosts(self, boosts: torch.Tensor):
        boosts = boosts + 6
        boosts1_embed = self.boosts1_onehot(boosts[..., :6]).squeeze(-1)
        boosts2_embed = self.boosts2_onehot(boosts[..., 6:]).squeeze(-1)
        return self.boosts_mlp(
            torch.cat((boosts1_embed, boosts2_embed), dim=-1).flatten(2)
        )

    def encode_volatiles(self, volatiles: torch.Tensor):
        return self.volatiles_mlp(volatiles.float().flatten(2))

    def encode_side_conditions(self, side_conditions: torch.Tensor):
        T, B, *_ = side_conditions.shape

        stealthrock = side_conditions[..., -4].unsqueeze(-1)
        stickyweb = side_conditions[..., -1].unsqueeze(-1)

        toxicspikes_onehot = self.toxicspikes_onehot(side_conditions[..., -3])
        spikes_onehot = self.spikes_onehot(side_conditions[..., -2])

        _, min_dur, max_dur = (
            side_conditions[..., :-4].view(T, B, 2, -1, 3).chunk(3, -1)
        )

        side_condition_embed = torch.cat(
            [
                stealthrock,
                stickyweb,
                toxicspikes_onehot,
                spikes_onehot,
                self.min_dur(min_dur).flatten(3),
                self.max_dur(max_dur).flatten(3),
            ],
            dim=-1,
        )
        return self.side_condt_mlp(side_condition_embed.flatten(2))

    def moves_given_context(self, active: torch.Tensor, moves: torch.Tensor):
        active = active.unsqueeze(-2)  # .repeat_interleave(4, -2)
        moves = moves[..., : self.n_active, :, :]
        return self.active_moves(active, moves)

    def forward(
        self,
        side: torch.Tensor,
        boosts: torch.Tensor,
        volatiles: torch.Tensor,
        side_conditions: torch.Tensor,
        wisher: torch.Tensor,
    ) -> SideEncoderOutput:

        pokemon_enc, mask, moves_emb = self.embedding.forward(side)

        T, B, S, *_ = pokemon_enc.shape

        mask = mask.clone()
        empty_mask = (mask).sum(-1) == 0
        if torch.any(empty_mask):
            mask[empty_mask, 0] = 0

        pokemon_embeddings = pokemon_enc.view(T * B, S, -1)
        mask = mask.view(T * B, S)

        pokemon_embeddings = self.transformer(pokemon_embeddings, mask)
        pokemon_embeddings = pokemon_embeddings * mask.unsqueeze(-1)

        pokemon_embedding = self.output(pokemon_embeddings)

        pokemon_embeddings = pokemon_embeddings.view(T, B, 3 * 12, -1)
        pokemon_embedding = pokemon_embedding.view(T, B, -1)

        boosts = self.encode_boosts(boosts)
        volatiles = self.encode_volatiles(volatiles)
        side_conditions = self.encode_side_conditions(side_conditions)

        active = pokemon_embeddings[..., : self.n_active, :]
        moves = self.moves_given_context(active, moves_emb[:, :, 0])
        switches = pokemon_embeddings[:, :, :6]

        return SideEncoderOutput(
            pokemon_embedding=pokemon_embedding,
            boosts=boosts,
            volatiles=volatiles,
            side_conditions=side_conditions,
            moves=moves,
            switches=switches,
        )



import torch
import torch.nn as nn
import torch.nn.functional as F

from meloetta.frameworks.porygon.model import config
from meloetta.frameworks.porygon.model.utils import _legal_policy

from meloetta.data import CHOICE_FLAGS


class FlagsHead(nn.Module):
    def __init__(self, config: config.FlagHeadConfig):
        super().__init__()

        self.mlp = nn.Sequential(
            nn.Linear(config.autoregressive_embedding_dim, config.hidden_dim),
            nn.LayerNorm(config.hidden_dim),
            nn.ReLU(),
            nn.Linear(config.hidden_dim, len(CHOICE_FLAGS)),
        )
        self.proj_flag = nn.Sequential(
            nn.Linear(len(CHOICE_FLAGS), config.hidden_dim),
            nn.LayerNorm(config.hidden_dim),
            nn.ReLU(),
            nn.Linear(config.hidden_dim, config.autoregressive_embedding_dim),
        )
        self.one_hot = nn.Embedding.from_pretrained(torch.eye(len(CHOICE_FLAGS)))

    def forward(
        self,
        action_index: torch.Tensor,
        autoregressive_embedding: torch.Tensor,
        flag_mask: torch.Tensor,
    ):
        T, B, *_ = autoregressive_embedding.shape

        flag_logits = self.mlp(autoregressive_embedding)
        flag_policy = _legal_policy(flag_logits, flag_mask)
        flag_index = torch.multinomial(flag_policy.view(T * B, -1), 1)
        flag_index = flag_index.view(T, B, -1)

        flag_one_hot = self.one_hot(flag_index.long())
        projected_flag_embedding = self.proj_flag(flag_one_hot)
        projected_flag_embedding = projected_flag_embedding.view(T, B, -1)

        valid_indices = action_index < 4
        valid_indices = valid_indices.view(T, B, 1)

        return (
            flag_logits,
            flag_policy,
            flag_index,
            autoregressive_embedding + valid_indices * projected_flag_embedding,
        )



import torch
import torch.nn as nn
import torch.nn.functional as F

from meloetta.frameworks.mewzero.model import config


class ResidualLayer(nn.Module):
    def __init__(self, channels: int) -> None:
        super().__init__()
        self.res = nn.Sequential(
            nn.Linear(channels, channels),
            nn.ReLU(),
            nn.Linear(channels, channels),
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return F.relu(x + self.res(x))


class PrivateSpatialEncoder(nn.Module):
    def __init__(self, gen: int, n_active: int, config: config.PrivateEncoderConfig):
        super().__init__()

        self.config = config
        self.gen = gen
        self.n_active = n_active

        self.project = nn.Linear(
            config.entity_embedding_dim * 2, config.entity_embedding_dim
        )

        self.resblock_stacks = nn.Sequential(
            *[ResidualLayer(config.entity_embedding_dim) for _ in range(2)]
        )

        self.spatial_out = nn.Linear(
            config.entity_embedding_dim, config.entity_embedding_dim
        )

    def forward(
        self, entity_embeddings: torch.Tensor, mask: torch.Tensor
    ) -> torch.Tensor:
        active_mask = mask[..., : self.n_active, :]
        active = entity_embeddings[..., : self.n_active, :].sum(1, keepdim=True)
        active = active / active_mask.sum(1, keepdim=True).clamp(min=1)

        reserve_mask = mask[..., self.n_active :, :]
        reserve = entity_embeddings[..., self.n_active :, :].sum(1, keepdim=True)
        reserve = reserve / reserve_mask.sum(1, keepdim=True).clamp(min=1)

        spatial_emb = torch.cat((active, reserve), dim=-1)
        spatial_emb = self.project(spatial_emb)
        spatial_emb = self.resblock_stacks(spatial_emb)
        spatial_emb = self.spatial_out(spatial_emb)

        return spatial_emb



import os
import json
import numpy as np

from copy import deepcopy
from typing import Dict, Any

ROOT_DIR = os.path.dirname(os.path.realpath(__file__))
DATA_DIR = f"{ROOT_DIR}/js/data"

with open(os.path.join(DATA_DIR, "BattleAbilities.json"), "r") as f:
    BattleAbilities = json.loads(f.read())

with open(os.path.join(DATA_DIR, "BattleAliases.json"), "r") as f:
    BattleAliases = json.loads(f.read())

with open(os.path.join(DATA_DIR, "BattleArticleTitles.json"), "r") as f:
    BattleArticleTitles = json.loads(f.read())

with open(os.path.join(DATA_DIR, "BattleFormatsData.json"), "r") as f:
    BattleFormatsData = json.loads(f.read())

with open(os.path.join(DATA_DIR, "BattleItems.json"), "r") as f:
    BattleItems = json.loads(f.read())

with open(os.path.join(DATA_DIR, "BattleLearnsets.json"), "r") as f:
    BattleLearnsets = json.loads(f.read())

with open(os.path.join(DATA_DIR, "BattleMovedex.json"), "r") as f:
    BattleMovedex = json.loads(f.read())

with open(os.path.join(DATA_DIR, "BattlePokedex.json"), "r") as f:
    BattlePokedex = json.loads(f.read())

with open(os.path.join(DATA_DIR, "BattleSearchCountIndex.json"), "r") as f:
    BattleSearchCountIndex = json.loads(f.read())

with open(os.path.join(DATA_DIR, "BattleSearchIndex.json"), "r") as f:
    BattleSearchIndex = json.loads(f.read())

with open(os.path.join(DATA_DIR, "BattleSearchIndexOffset.json"), "r") as f:
    BattleSearchIndexOffset = json.loads(f.read())

with open(os.path.join(DATA_DIR, "BattleTeambuilderTable.json"), "r") as f:
    BattleTeambuilderTable = json.loads(f.read())

with open(os.path.join(DATA_DIR, "BattleText.json"), "r") as f:
    BattleText = json.loads(f.read())

with open(os.path.join(DATA_DIR, "BattleTypeChart.json"), "r") as f:
    BattleTypeChart = json.loads(f.read())

with open(os.path.join(DATA_DIR, "Formats.json"), "r") as f:
    Formats = json.loads(f.read())


GMAX_MOVES = [move for move in BattleMovedex if "gmax" in move]


BOOSTS = ["atk", "def", "spc", "spa", "spd", "spe", "evasion", "accuracy"]


def load_feature_embedding(type: str, gen: int):
    return np.load(os.path.join("pretrained", f"gen{gen}", type + ".npy"))


Schema = Dict[str, Dict[str, Dict[str, Dict[str, Any]]]]
with open("meloetta/pretrained/schema.json", "r") as f:
    schema: Schema = json.loads(f.read())

TOKENIZED_SCHEMA = deepcopy(schema)


def to_id(value: Any):
    if isinstance(value, str):
        return "".join([c for c in value if c.isalnum()]).lower()
    else:
        return str(value)


for gen in TOKENIZED_SCHEMA:
    for dex_type in schema[gen]:
        for key, values in sorted(schema[gen][dex_type].items()):
            TOKENIZED_SCHEMA[gen][dex_type][key] = {
                to_id(value): index
                for index, value in enumerate(values)
                if to_id(value)
            }


def get_type_token(gen: int, value: Any):
    value = to_id(value)
    lookup = TOKENIZED_SCHEMA[f"gen{gen}"]["movedex"]["type"]
    return lookup.get(value, -1)


def get_species_token(gen: int, key: int, value: Any):
    value = to_id(value)
    lookup = TOKENIZED_SCHEMA[f"gen{gen}"]["pokedex"][key]
    return lookup.get(value, -1)


def get_move_token(gen: int, key: int, value: Any):
    value = to_id(value)
    lookup = TOKENIZED_SCHEMA[f"gen{gen}"]["movedex"][key]
    return lookup.get(value, -1)


def get_ability_token(gen: int, key: int, value: Any):
    value = to_id(value)
    lookup = TOKENIZED_SCHEMA[f"gen{gen}"]["abilitydex"][key]
    return lookup.get(value, -1)


def get_item_token(gen: int, key: int, value: Any):
    value = to_id(value)
    lookup = TOKENIZED_SCHEMA[f"gen{gen}"]["itemdex"][key]
    return lookup.get(value, -1)


GENDERS = {"M": 0, "F": 1, "N": 2}
STATUS = {"par": 0, "psn": 1, "frz": 2, "slp": 3, "brn": 4}


def get_gender_token(value: str):
    return GENDERS.get(value, -1)


def get_status_token(value: str):
    return STATUS.get(value, -1)


with open(f"{ROOT_DIR}/pretrained/wsnc.json", "r") as f:
    WSNC = json.loads(f.read())

VOLATILES = WSNC["volatiles"]
VOLATILES = {v: i for i, v in enumerate(VOLATILES)}

WEATHERS = WSNC["weathers"]
WEATHERS = {v: i for i, v in enumerate(WEATHERS)}

PSEUDOWEATHERS = WSNC["pseudoweather"]
TERRAIN = WSNC["terrain"]
PSEUDOWEATHERS = {v: i for i, v in enumerate(PSEUDOWEATHERS + TERRAIN)}

ITEM_EFFECTS = WSNC["item_effects"]
ITEM_EFFECTS = {v: i for i, v in enumerate(ITEM_EFFECTS) if v}


SIDE_CONDITIONS = WSNC["side_conditions"]
SIDE_CONDITIONS = {v: i for i, v in enumerate(SIDE_CONDITIONS) if v}


def get_item_effect_token(name: str):
    return ITEM_EFFECTS.get(name, -1)


def get_weather_token(name: str):
    return WEATHERS.get(name, -1)


def get_pseudoweather_token(name: str):
    return PSEUDOWEATHERS.get(name, -1)


def get_side_condition_token(name: str):
    return SIDE_CONDITIONS.get(name, -1)


# Choice related

CHOICE_TOKENS = {
    "move": 0,
    "switch": 1,
}

CHOICE_FLAGS = {
    "mega": 0,
    "zmove": 1,
    "dynamax": 2,
    "max": 2,
    "terastallize": 3,
}

CHOICE_TARGETS = list(range(-3, 3))
CHOICE_TARGETS.remove(0)
CHOICE_TARGETS = {str(target): i for i, target in enumerate(CHOICE_TARGETS)}


def get_choice_flag_token(name: str):
    return CHOICE_FLAGS[name]


def get_choice_target_token(name: str):
    return CHOICE_TARGETS[name]


def get_choice_token(name: str):
    return CHOICE_TOKENS[name]


_STATE_FIELDS = {
    "sides",
    "boosts",
    "volatiles",
    "side_conditions",
    "pseudoweathers",
    "weather",
    "wisher",
    "scalars",
    "prev_choices",
    "choices_done",
    "action_type_mask",
    "move_mask",
    "max_move_mask",
    "switch_mask",
    "flag_mask",
    "target_mask",
}



import json
import torch
import traceback

from torch.nn.utils.rnn import pad_sequence

from typing import Tuple, Dict, Any

from meloetta.room import BattleRoom

from meloetta.frameworks.porygon import ReplayBuffer
from meloetta.frameworks.porygon.model import (
    PorygonModel,
    ModelOutput,
    PostProcess,
)

from meloetta.actors.base import Actor
from meloetta.actors.types import State, Choices, Battle
from meloetta.utils import expand_bt
from meloetta.data import (
    BOOSTS,
    VOLATILES,
    PSEUDOWEATHERS,
    SIDE_CONDITIONS,
    get_item_effect_token,
    get_status_token,
    get_gender_token,
    get_species_token,
    get_ability_token,
    get_item_token,
    get_move_token,
    get_type_token,
    get_weather_token,
)

SIDES = ["mySide", "farSide"]


class PorygonActor(Actor):
    def __init__(
        self,
        model: PorygonModel,
        replay_buffer: ReplayBuffer = None,
        pid: str = None,
    ):
        self.model = model
        self.gen = model.gen

        self.replay_buffer = replay_buffer
        self.hidden_state = model.core.initial_state(1)

        if replay_buffer is not None:
            self.buffer_index = self.replay_buffer._get_index()
        self.step_index = 0

        self.env_outputs = []
        self.model_outputs = []

        self.username = pid

    @property
    def storing_transition(self):
        return self.replay_buffer is not None

    def choose_action(
        self,
        env_output: State,
        room: BattleRoom,
        choices: Choices,
    ):
        output: Tuple[ModelOutput, PostProcess]
        with torch.no_grad():
            output = self.model(
                env_output, self.hidden_state, choices, battle_tag=room.battle_tag
            )
        model_output, postprocess, self.hidden_state = output

        if self.storing_transition:
            self._store_transition(env_output, model_output, room)

        data = postprocess.data
        index = postprocess.index
        try:
            func, args, kwargs = data[index]
        except Exception:
            traceback.print_exc()

            battle_tag = room.battle_tag
            datum = {
                "ModelOutput": {
                    esk: {k: v for k, v in esv._asdict().items() if v is not None}
                    if not isinstance(esv, torch.Tensor)
                    else esv
                    for esk, esv in model_output._asdict().items()
                    if esv is not None
                },
                "index": index,
            }
            with open(f"errors/{battle_tag}-nashketchum_model.json", "w") as f:
                json.dump(datum, f)

            return self.choose_action(env_output, room, choices)
        else:
            if self.storing_transition:
                self._store_transition(env_output, model_output, room)

        return func, args, kwargs

    def _clean_env_output(self, env_output: State):
        return {
            k: env_output[k]
            for k in self.model.state_fields
            if isinstance(env_output[k], torch.Tensor)
        }

    def _clean_model_output(self, model_output: ModelOutput):
        to_store = {}
        to_store.update(
            {
                k: v.squeeze()
                for k, v in model_output.indices._asdict().items()
                if isinstance(v, torch.Tensor)
            }
        )
        to_store.update(
            {
                k: v.squeeze()
                for k, v in model_output._asdict().items()
                if isinstance(v, torch.Tensor)
            }
        )
        return to_store

    def _store_transition(
        self, env_output: State, model_output: ModelOutput, room: BattleRoom
    ):
        env_output = self._clean_env_output(env_output)
        model_output = self._clean_model_output(model_output)

        self.env_outputs.append(env_output)
        self.model_outputs.append(model_output)

    def _populate_buffer(self) -> int:
        for step_index, (env_output, model_output) in enumerate(
            zip(self.env_outputs, self.model_outputs)
        ):
            to_store = {**env_output, **model_output}

            self.replay_buffer.store_sample(self.buffer_index, step_index, to_store)

        return step_index

    def post_match(self, room: BattleRoom):
        if self.storing_transition:
            reward = room.get_js_attr("reward")

            final_turn = self._populate_buffer()

            self.replay_buffer.append_reward(self.buffer_index, final_turn, reward)

    def _vectorize_pokemon(self, pokemon: Dict[str, Any], active: bool) -> torch.Tensor:
        return pokemon

    def _vectorize_pokemon(
        self, pokemon: Dict[str, Any], public: bool = False
    ) -> torch.Tensor:
        moves = pokemon["moves"]
        hp = pokemon.get("hp", 0) * (not public)
        maxhp = pokemon.get("maxhp", 0) * (not public)
        hp_ratio = pokemon.get("hp", 0) / max(pokemon.get("maxhp", 0), 1)
        fainted = hp_ratio == 0
        return torch.tensor(
            [
                get_species_token(self.gen, "name", pokemon["name"]),
                get_species_token(self.gen, "forme", pokemon["speciesForme"]),
                pokemon.get("slot", 0),
                hp,
                maxhp,
                hp_ratio,
                pokemon.get("stats", {}).get("atk", -1),
                pokemon.get("stats", {}).get("def", -1),
                pokemon.get("stats", {}).get("spa", -1),
                pokemon.get("stats", {}).get("spd", -1),
                pokemon.get("stats", {}).get("spe", -1),
                fainted,
                pokemon.get("active", 0),
                pokemon.get("level", 0),
                get_gender_token(pokemon.get("gender", "")),
                get_ability_token(self.gen, "name", pokemon["ability"]),
                get_ability_token(self.gen, "name", pokemon["baseAbility"]),
                get_item_token(self.gen, "name", pokemon.get("item", "")),
                get_item_token(self.gen, "name", pokemon.get("prevItem", "")),
                get_item_effect_token(pokemon.get("itemEffect", "")),
                get_item_effect_token(pokemon.get("prevItemEffect", "")),
                get_status_token(pokemon.get("status", "")),
                min(pokemon.get("statusData", {}).get("sleepTurns", 0), 3),
                min(pokemon.get("statusData", {}).get("toxicTurns", 0), 15),
                get_move_token(self.gen, "id", pokemon.get("lastMove")),
                get_move_token(
                    self.gen, "id", moves[min(0, len(moves) - 1)] if moves else ""
                ),
                get_move_token(
                    self.gen, "id", moves[min(1, len(moves) - 1)] if moves else ""
                ),
                get_move_token(
                    self.gen, "id", moves[min(2, len(moves) - 1)] if moves else ""
                ),
                get_move_token(
                    self.gen, "id", moves[min(3, len(moves) - 1)] if moves else ""
                ),
                get_type_token(self.gen, pokemon.get("terastallized")),
                get_type_token(self.gen, pokemon.get("teraType")),
                min(pokemon.get("timesAttacked", 0), 6),
            ]
        )

    def get_vectorized_state(
        self, room: BattleRoom, battle: Battle
    ) -> Dict[str, torch.Tensor]:
        my_private_side = {p["searchid"]: p for p in battle["myPokemon"]}

        my_public_side = {
            p["searchid"]: p for p in battle["mySide"]["pokemon"] if p is not None
        }
        my_active = {p["searchid"] for p in battle["mySide"]["active"] if p is not None}

        opp_public_side = {
            p["searchid"]: p for p in battle["farSide"]["pokemon"] if p is not None
        }
        opp_active = {
            p["searchid"] for p in battle["farSide"]["active"] if p is not None
        }

        boosts = [None, None]
        volatiles = [None, None]

        for sid, (public_side, active_list) in enumerate(
            zip([my_public_side, opp_public_side], [my_active, opp_active])
        ):
            for pokemon in public_side.values():
                if pokemon["searchid"] in active_list:
                    boosts_vector = torch.tensor(
                        [pokemon["boosts"].get(boost, 0) for boost in BOOSTS]
                    )
                    boosts[sid] = boosts_vector

                    volatile_vector = torch.tensor(
                        [
                            1 if volatile in pokemon["volatiles"] else 0
                            for volatile in VOLATILES
                        ]
                    )
                    volatiles[sid] = volatile_vector

        for sid in range(len(boosts)):
            if boosts[sid] is None:
                boosts[sid] = torch.zeros(len(BOOSTS), dtype=torch.long)

        for sid in range(len(volatiles)):
            if volatiles[sid] is None:
                volatiles[sid] = torch.zeros(len(VOLATILES), dtype=torch.long)

        boosts = torch.stack(boosts)
        volatiles = torch.stack(volatiles)

        side_conditions = []
        n = []
        total_pokemon = []
        faint_counter = []
        wisher_slot = []

        for side in SIDES:
            side_datum = battle[side]
            sc_datum = battle[side]["sideConditions"]

            side_conditions_vector = [
                sc_datum.get(side_condition, [None, 0, 0, 0])[1:]
                for side_condition in SIDE_CONDITIONS
                if side_condition
                not in {"stealthrock", "spikes", "toxicspikes", "stickyweb"}
            ]
            side_conditions_vector = [i for o in side_conditions_vector for i in o]
            side_conditions_vector += [sc_datum.get("stealthrock", [None, 0])[1]]
            side_conditions_vector += [sc_datum.get("toxicspikes", [None, 0])[1]]
            side_conditions_vector += [sc_datum.get("spikes", [None, 0])[1]]
            side_conditions_vector += [sc_datum.get("stickyweb", [None, 0])[1]]
            side_conditions += [torch.tensor(side_conditions_vector)]

            n += [torch.tensor(side_datum["n"])]
            total_pokemon += [torch.tensor(side_datum["totalPokemon"])]
            faint_counter += [torch.tensor(min(6, side_datum["faintCounter"]))]
            wisher_slot += [
                torch.tensor((side_datum["wisher"] or {"slot": -1})["slot"])
            ]

        side_conditions = torch.stack(side_conditions)
        n = torch.stack(n)
        total_pokemon = torch.stack(total_pokemon)
        faint_counter = torch.stack(faint_counter)
        wisher_slot = torch.stack(wisher_slot)

        my_private_side_vectors = []
        for i, (sid, private_data) in enumerate(my_private_side.items()):
            datum = {}

            public_data = my_public_side.get(sid, {})
            if i == 0:
                keys = list(public_data) + list(private_data)

            datum["active"] = sid in my_active
            for key in keys:
                value = private_data.get(key) or public_data.get(key)
                if value is not None:
                    datum[key] = value

            private_vector = self._vectorize_pokemon(datum)
            my_private_side_vectors.append(private_vector)

        my_public_side_vectors = []
        for i, (sid, public_data) in enumerate(my_public_side.items()):
            datum = {}
            datum["active"] = sid in my_active
            for key in keys:
                if key in public_data:
                    value = public_data.get(key)
                    if value is not None:
                        datum[key] = value

            public_vector = self._vectorize_pokemon(datum, public=True)
            my_public_side_vectors.append(public_vector)

        opp_public_side_vectors = []
        for i, (sid, public_data) in enumerate(opp_public_side.items()):
            datum = {}
            datum["active"] = sid in opp_active
            for key in keys:
                if key in public_data:
                    value = public_data.get(key)
                    if value is not None:
                        datum[key] = value

            public_vector = self._vectorize_pokemon(datum, public=True)
            opp_public_side_vectors.append(public_vector)

        my_private_side = torch.stack(my_private_side_vectors)
        my_private_side = torch.cat(
            (my_private_side, -torch.ones_like(my_private_side))
        )

        my_public_side = torch.stack(my_public_side_vectors)
        opp_public_side = torch.stack(opp_public_side_vectors)

        sides = pad_sequence(
            [my_private_side, my_public_side, opp_public_side],
            batch_first=True,
            padding_value=-1,
        )

        pseudoweathers = {
            (pseudoweather[0]): pseudoweather[1:]
            for pseudoweather in battle["pseudoWeather"]
        }
        pseudoweathers = [
            torch.tensor(pseudoweathers.get(pseudoweather, [-1, -1]))
            for pseudoweather in PSEUDOWEATHERS
        ]
        pseudoweathers = torch.stack(pseudoweathers)

        weather = torch.tensor(
            [
                get_weather_token(battle["weather"]),
                battle["weatherTimeLeft"],
                battle["weatherMinTimeLeft"],
            ]
        )

        turn = torch.tensor(battle["turn"])

        state = {
            "sides": sides,
            "boosts": boosts,
            "volatiles": volatiles,
            "side_conditions": side_conditions,
            "pseudoweathers": pseudoweathers,
            "weather": weather,
            "wisher": wisher_slot,
            "turn": turn,
            "n": n,
            "total_pokemon": total_pokemon,
            "faint_counter": faint_counter,
        }
        state = {key: expand_bt(value) for key, value in state.items()}
        return state



from meloetta.frameworks.mewzero.model.heads.action_type_head import ActionTypeHead
from meloetta.frameworks.mewzero.model.heads.max_move_head import MaxMoveHead
from meloetta.frameworks.mewzero.model.heads.move_head import MoveHead
from meloetta.frameworks.mewzero.model.heads.switch_head import SwitchHead
from meloetta.frameworks.mewzero.model.heads.flag_head import FlagsHead



import torch
import random
import threading
import multiprocessing as mp

from typing import Dict, List

from meloetta.frameworks.porygon.model.interfaces import Batch
from meloetta.frameworks.porygon.utils import create_buffers


Buffers = Dict[str, List[torch.Tensor]]


class ReplayBuffer:
    def __init__(
        self,
        trajectory_length: int,
        gen: int,
        gametype: str,
        num_buffers: int = 512,
        device: torch.device = torch.device(
            "cuda" if torch.cuda.is_available() else "cpu"
        ),
    ):
        self.num_buffers = num_buffers
        self.device = device
        self.finish_queue = mp.Queue()

        manager = mp.Manager()
        self.index_cache = manager.dict()

        self.buffers = create_buffers(num_buffers, trajectory_length, gen, gametype)

        self.valid_masks = [field for field in self.buffers if field.endswith("_mask")]

        self.full_queue = mp.Queue()
        self.free_queue = mp.Queue()

        for m in range(num_buffers):
            self.free_queue.put(m)

    def _get_index(self) -> int:
        index = self.free_queue.get()
        self._reset_index(index)
        return index

    def store_sample(
        self,
        buffer_index: int,
        turn_index: int,
        step: Dict[str, torch.Tensor],
    ):
        for key, value in step.items():
            try:
                self.buffers[key][buffer_index][turn_index][...] = value
            except Exception as e:
                raise e
        self.buffers["valid"][buffer_index][turn_index][...] = 1

    def append_reward(self, buffer_index: int, turn_index: int, reward: int):
        self.buffers["rewards"][buffer_index][turn_index][...] = reward
        length = sum(self.buffers["valid"][buffer_index]).item()
        self.finish_queue.put((None, length))
        self.full_queue.put(buffer_index)

    def _reset_index(self, index: int):
        self.buffers["valid"][index][...] = 0
        self.buffers["rewards"][index][...] = 0
        for valid_mask in self.valid_masks:
            self.buffers[valid_mask][index][...] = 1

    def get_batch(
        self,
        batch_size: int,
        lock=threading.Lock(),
    ) -> Batch:
        with lock:
            indices = [self.full_queue.get() for _ in range(batch_size)]

        valids = torch.stack([self.buffers["valid"][m] for m in indices])
        lengths = valids.sum(-1)

        if torch.any(lengths == 0):
            return None

        max_length = lengths.max().item()

        indices = list(
            zip(
                *sorted(
                    list(zip(lengths.tolist(), indices)),
                    key=lambda x: x[0],
                    reverse=True,
                )
            )
        )[1]

        batch = {
            key: torch.stack(
                [self.buffers[key][index][:max_length] for index in indices],
                dim=1,
            )
            for key in self.buffers
        }

        for m in indices:
            self.free_queue.put(m)

        batch = {
            k: t.to(device=self.device, non_blocking=True) for k, t in batch.items()
        }
        return Batch(**batch)



import re
import torch

from typing import Dict, List, Any, Sequence, Callable, NamedTuple, Tuple


from meloetta.data import CHOICE_FLAGS


class LoopVTraceCarry(NamedTuple):
    """The carry of the v-trace scan loop."""

    reward: torch.Tensor
    # The cumulated reward until the end of the episode. Uncorrected (v-trace).
    # Gamma discounted and includes eta_reg_entropy.
    reward_uncorrected: torch.Tensor
    next_value: torch.Tensor
    next_v_target: torch.Tensor
    importance_sampling: torch.Tensor


def _player_others(
    player_ids: torch.Tensor, valid: torch.Tensor, player: int
) -> torch.Tensor:
    """A vector of 1 for the current player and -1 for others.

    Args:
      player_ids: Tensor [...] containing player ids (0 <= player_id < N).
      valid: Tensor [...] containing whether these states are valid.
      player: The player id as int.

    Returns:
      player_other: is 1 for the current player and -1 for others [..., 1].
    """
    current_player_tensor = player_ids == player

    res = 2 * current_player_tensor - 1
    res = res * valid
    return torch.unsqueeze(res, dim=-1)


def _where(
    pred: torch.Tensor, true_data: torch.Tensor, false_data: torch.Tensor
) -> torch.Tensor:
    """Similar to jax.where but treats `pred` as a broadcastable prefix."""

    def _where_one(t, f):
        if (
            isinstance(t, LoopVTraceCarry)
            or isinstance(t, tuple)
            or isinstance(t, list)
        ):
            res = [_where_one(ts, fs) for ts, fs in zip(t, f)]
            if isinstance(t, LoopVTraceCarry):
                return LoopVTraceCarry(*res)
            else:
                return res
        else:
            # Expand the dimensions of pred if true_data and false_data are higher rank.
            p = pred.view(pred.shape + (1,) * (len(t.shape) - len(pred.shape)))
            p = p.to(torch.bool)
            return torch.where(p, t, f)

    output = [_where_one(td, fd) for td, fd in zip(true_data, false_data)]
    return output


def scan(f: Callable, init, xs, length=None, reverse: bool = False):
    if xs is None:
        xs = [None] * length
    carry = init
    ys = []

    if reverse:
        xs = list(reversed(list(zip(*xs))))

    for x in xs:
        carry, y = f(carry, x)
        ys.append(y)

    if reverse:
        ys = list(reversed(ys))

    if isinstance(ys[0], list):
        res = [torch.stack([n[i] for n in ys]) for i in range(len(ys[0]))]
    else:
        res = torch.stack(ys)

    return carry, res


def _has_played(
    valid: torch.Tensor, player_id: torch.Tensor, player: int
) -> torch.Tensor:
    """Compute a mask of states which have a next state in the sequence."""
    assert valid.shape == player_id.shape

    def _loop_has_played(carry, x):
        valid, player_id = x
        assert valid.shape == player_id.shape

        our_res = torch.ones_like(player_id)
        opp_res = carry
        reset_res = torch.zeros_like(carry)

        our_carry = carry
        opp_carry = carry
        reset_carry = torch.zeros_like(player_id)

        # pyformat: disable
        return _where(
            valid,
            _where(
                (player_id == player),
                (our_carry, our_res),
                (opp_carry, opp_res),
            ),
            (reset_carry, reset_res),
        )
        # pyformat: enable

    _, result = scan(
        f=_loop_has_played,
        init=torch.zeros_like(player_id[-1]),
        xs=(valid, player_id),
        reverse=True,
    )
    return result


def _policy_ratio(
    pi: torch.Tensor, mu: torch.Tensor, actions_oh: torch.Tensor, valid: torch.Tensor
) -> torch.Tensor:
    """Returns a ratio of policy pi/mu when selecting action a.

    By convention, this ratio is 1 on non valid states
    Args:
      pi: the policy of shape [..., A].
      mu: the sampling policy of shape [..., A].
      actions_oh: a one-hot encoding of the current actions of shape [..., A].
      valid: 0 if the state is not valid and else 1 of shape [...].

    Returns:
      pi/mu on valid states and 1 otherwise. The shape is the same
      as pi, mu or actions_oh but without the last dimension A.
    """
    assert pi.shape == mu.shape == actions_oh.shape
    # assert ((valid,), actions_oh.shape[:-1])

    def _select_action_prob(pi: torch.Tensor) -> torch.Tensor:
        return torch.sum(actions_oh * pi, dim=-1) * valid + ~valid

    pi_actions_prob = _select_action_prob(pi)
    mu_actions_prob = _select_action_prob(mu)
    return pi_actions_prob / mu_actions_prob


@torch.no_grad()
def v_trace(
    v: torch.Tensor,
    valid: torch.Tensor,
    player_id: torch.Tensor,
    acting_policy: torch.Tensor,
    merged_policy: torch.Tensor,
    merged_log_policy: torch.Tensor,
    player_others: torch.Tensor,
    actions_oh: torch.Tensor,
    reward: torch.Tensor,
    player: int,
    # Scalars below.
    eta: float,
    lambda_: float,
    c: float,
    rho: float,
    gamma: float = 1.0,
) -> Tuple[Any, Any, Any]:
    """Custom VTrace for trajectories with a mix of different player steps."""

    has_played = _has_played(valid, player_id, player)

    policy_ratio = _policy_ratio(merged_policy, acting_policy, actions_oh, valid)
    inv_mu = _policy_ratio(
        torch.ones_like(merged_policy), acting_policy, actions_oh, valid
    )

    eta_reg_entropy = (
        -eta
        * torch.sum(merged_policy * merged_log_policy, dim=-1)
        * torch.squeeze(player_others, dim=-1)
    )
    eta_log_policy = -eta * merged_log_policy * player_others

    init_state_v_trace = LoopVTraceCarry(
        reward=torch.zeros_like(reward[-1]),
        reward_uncorrected=torch.zeros_like(reward[-1]),
        next_value=torch.zeros_like(v[-1]),
        next_v_target=torch.zeros_like(v[-1]),
        importance_sampling=torch.ones_like(policy_ratio[-1]),
    )

    def _loop_v_trace(carry: LoopVTraceCarry, x) -> Tuple[LoopVTraceCarry, Any]:
        (
            cs,
            player_id,
            v,
            reward,
            eta_reg_entropy,
            valid,
            inv_mu,
            actions_oh,
            eta_log_policy,
        ) = x

        reward_uncorrected = reward + gamma * carry.reward_uncorrected + eta_reg_entropy
        discounted_reward = reward + gamma * carry.reward

        # V-target:
        our_v_target = (
            v
            + torch.unsqueeze(
                torch.clamp(cs * carry.importance_sampling, max=rho), dim=-1
            )
            * (
                torch.unsqueeze(reward_uncorrected, dim=-1)
                + gamma * carry.next_value
                - v
            )
            + lambda_
            * torch.unsqueeze(
                torch.clamp(cs * carry.importance_sampling, max=c), dim=-1
            )
            * gamma
            * (carry.next_v_target - carry.next_value)
        )

        opp_v_target = torch.zeros_like(our_v_target)
        reset_v_target = torch.zeros_like(our_v_target)

        # Learning output:
        our_learning_output = (
            v
            + eta_log_policy  # value
            + actions_oh  # regularisation
            * torch.unsqueeze(inv_mu, dim=-1)
            * (
                torch.unsqueeze(discounted_reward, dim=-1)
                + gamma
                * torch.unsqueeze(carry.importance_sampling, dim=-1)
                * carry.next_v_target
                - v
            )
        )

        opp_learning_output = torch.zeros_like(our_learning_output)
        reset_learning_output = torch.zeros_like(our_learning_output)

        # State carry:
        our_carry = LoopVTraceCarry(
            reward=torch.zeros_like(carry.reward),
            next_value=v,
            next_v_target=our_v_target,
            reward_uncorrected=torch.zeros_like(carry.reward_uncorrected),
            importance_sampling=torch.ones_like(carry.importance_sampling),
        )
        opp_carry = LoopVTraceCarry(
            reward=eta_reg_entropy + cs * discounted_reward,
            reward_uncorrected=reward_uncorrected,
            next_value=gamma * carry.next_value,
            next_v_target=gamma * carry.next_v_target,
            importance_sampling=cs * carry.importance_sampling,
        )
        reset_carry = init_state_v_trace

        # Invalid turn: init_state_v_trace and (zero target, learning_output)
        # pyformat: disable
        return _where(
            valid,
            _where(
                (player_id == player),
                (our_carry, (our_v_target, our_learning_output)),
                (opp_carry, (opp_v_target, opp_learning_output)),
            ),
            (reset_carry, (reset_v_target, reset_learning_output)),
        )
        # pyformat: enable

    _, (v_target, learning_output) = scan(
        f=_loop_v_trace,
        init=init_state_v_trace,
        xs=(
            policy_ratio,
            player_id,
            v,
            reward,
            eta_reg_entropy,
            valid,
            inv_mu,
            actions_oh,
            eta_log_policy,
        ),
        reverse=True,
    )

    return v_target, has_played, learning_output


def apply_force_with_threshold(
    decision_outputs: torch.Tensor,
    force: torch.Tensor,
    threshold: float,
    threshold_center: torch.Tensor,
) -> torch.Tensor:
    """Apply the force with below a given threshold."""
    can_decrease = decision_outputs - threshold_center > -threshold
    can_increase = decision_outputs - threshold_center < threshold
    force_negative = torch.clamp(force, max=0.0)
    force_positive = torch.clamp(force, min=0.0)
    clipped_force = can_decrease * force_negative + can_increase * force_positive
    return decision_outputs * clipped_force.detach()


def renormalize(loss: torch.Tensor, mask: torch.Tensor) -> torch.Tensor:
    """The `normalization` is the number of steps over which loss is computed."""
    loss = torch.sum(loss * mask)
    return loss


def get_loss_v(
    v_list: Sequence[torch.Tensor],
    v_target_list: Sequence[torch.Tensor],
    mask_list: Sequence[torch.Tensor],
) -> torch.Tensor:
    """Define the loss function for the critic."""
    loss_v_list = []
    for (v_n, v_target, mask) in zip(v_list, v_target_list, mask_list):
        assert v_n.shape[0] == v_target.shape[0]

        loss_v = torch.unsqueeze(mask, dim=-1) * (v_n - v_target.detach()) ** 2
        loss_v = torch.sum(loss_v)
        loss_v_list.append(loss_v)

    return sum(loss_v_list)


def get_loss_nerd(
    logit_list: Sequence[torch.Tensor],
    policy_list: Sequence[torch.Tensor],
    q_vr_list: Sequence[torch.Tensor],
    valid: torch.Tensor,
    player_ids: Sequence[torch.Tensor],
    legal_actions: torch.Tensor,
    importance_sampling_correction: Sequence[torch.Tensor],
    clip: float = 100,
    threshold: float = 2,
) -> torch.Tensor:
    """Define the nerd loss."""
    assert isinstance(importance_sampling_correction, list)
    loss_pi_list = []
    for k, (logit_pi, pi, q_vr, is_c) in enumerate(
        zip(logit_list, policy_list, q_vr_list, importance_sampling_correction)
    ):
        assert logit_pi.shape[0] == q_vr.shape[0]
        # loss policy
        adv_pi = q_vr - torch.sum(pi * q_vr, dim=-1, keepdim=True)
        adv_pi = is_c * adv_pi  # importance sampling correction
        adv_pi = torch.clip(adv_pi, min=-clip, max=clip)
        adv_pi = adv_pi.detach()

        logits = logit_pi - torch.mean(logit_pi * legal_actions, dim=-1, keepdim=True)

        threshold_center = torch.zeros_like(logits)

        force_threshold = apply_force_with_threshold(
            logits,
            adv_pi,
            threshold,
            threshold_center,
        )
        nerd_loss = torch.sum(legal_actions * force_threshold, axis=-1)
        nerd_loss = -renormalize(nerd_loss, valid * (player_ids == k))
        loss_pi_list.append(nerd_loss)

    return sum(loss_pi_list)


def get_gen_and_gametype(battle_format: str) -> Tuple[int, str]:
    gen = int(re.search(r"gen([0-9])", battle_format).groups()[0])
    if "triples" in battle_format:
        gametype = "triples"
    if "doubles" in battle_format:
        gametype = "doubles"
    else:
        gametype = "singles"
    return gen, gametype


def to_id(string: str):
    return "".join(c for c in string if c.isalnum()).lower()


def _get_private_reserve_size(gen: int):
    if gen == 9:
        return 28
    elif gen == 8:
        return 25
    else:
        return 24


def get_buffer_specs(
    trajectory_length: int,
    gen: int,
    gametype: str,
    private_reserve_size: int,
):
    if gametype != "singles":
        if gametype == "doubles":
            n_active = 2
        else:
            n_active = 3
    else:
        n_active = 1

    buffer_specs = {
        "private_reserve": {
            "size": (trajectory_length, 6, private_reserve_size),
            "dtype": torch.int64,
        },
        "public_n": {
            "size": (trajectory_length, 2),
            "dtype": torch.int64,
        },
        "public_total_pokemon": {
            "size": (trajectory_length, 2),
            "dtype": torch.int64,
        },
        "public_faint_counter": {
            "size": (trajectory_length, 2),
            "dtype": torch.int64,
        },
        "public_side_conditions": {
            "size": (trajectory_length, 2, 18, 3),
            "dtype": torch.int64,
        },
        "public_wisher": {
            "size": (trajectory_length, 2),
            "dtype": torch.int64,
        },
        "public_active": {
            "size": (trajectory_length, 2, n_active, 158),
            "dtype": torch.int64,
        },
        "public_reserve": {
            "size": (trajectory_length, 2, 6, 37),
            "dtype": torch.int64,
        },
        "public_stealthrock": {
            "size": (trajectory_length, 2),
            "dtype": torch.int64,
        },
        "public_spikes": {
            "size": (trajectory_length, 2),
            "dtype": torch.int64,
        },
        "public_toxicspikes": {
            "size": (trajectory_length, 2),
            "dtype": torch.int64,
        },
        "public_stickyweb": {
            "size": (trajectory_length, 2),
            "dtype": torch.int64,
        },
        "weather": {
            "size": (trajectory_length,),
            "dtype": torch.int64,
        },
        "weather_time_left": {
            "size": (trajectory_length,),
            "dtype": torch.int64,
        },
        "weather_min_time_left": {
            "size": (trajectory_length,),
            "dtype": torch.int64,
        },
        "pseudo_weather": {
            "size": (trajectory_length, 12, 2),
            "dtype": torch.int64,
        },
        "turn": {
            "size": (trajectory_length,),
            "dtype": torch.int64,
        },
        "turns_since_last_move": {
            "size": (trajectory_length,),
            "dtype": torch.int64,
        },
        "action_type_mask": {
            "size": (trajectory_length, 3),
            "dtype": torch.bool,
        },
        "move_mask": {
            "size": (trajectory_length, 4),
            "dtype": torch.bool,
        },
        "switch_mask": {
            "size": (trajectory_length, 6),
            "dtype": torch.bool,
        },
        "flag_mask": {
            "size": (trajectory_length, len(CHOICE_FLAGS)),
            "dtype": torch.bool,
        },
        "rewards": {
            "size": (trajectory_length, 2),
            "dtype": torch.float32,
        },
        "player_id": {
            "size": (trajectory_length,),
            "dtype": torch.long,
        },
    }
    if gametype != "singles":
        buffer_specs.update(
            {
                "target_mask": {
                    "size": (trajectory_length, 4),
                    "dtype": torch.bool,
                },
                "prev_choices": {
                    "size": (trajectory_length, 2, 4),
                    "dtype": torch.int64,
                },
                "choices_done": {
                    "size": (trajectory_length, 1),
                    "dtype": torch.int64,
                },
                "targeting": {
                    "size": (trajectory_length, 1),
                    "dtype": torch.int64,
                },
                "target_policy": {
                    "size": (trajectory_length, 2 * n_active),
                    "dtype": torch.float32,
                },
                "target_index": {
                    "size": (trajectory_length,),
                    "dtype": torch.int64,
                },
            }
        )

    # add policy...
    buffer_specs.update(
        {
            "action_type_policy": {
                "size": (trajectory_length, 3),
                "dtype": torch.float32,
            },
            "move_policy": {
                "size": (trajectory_length, 4),
                "dtype": torch.float32,
            },
            "switch_policy": {
                "size": (trajectory_length, 6),
                "dtype": torch.float32,
            },
            "flag_policy": {
                "size": (trajectory_length, len(CHOICE_FLAGS)),
                "dtype": torch.float32,
            },
        }
    )

    # ...and indices
    buffer_specs.update(
        {
            "action_type_index": {
                "size": (trajectory_length,),
                "dtype": torch.int64,
            },
            "move_index": {
                "size": (trajectory_length,),
                "dtype": torch.int64,
            },
            "switch_index": {
                "size": (trajectory_length,),
                "dtype": torch.int64,
            },
            "flag_index": {
                "size": (trajectory_length,),
                "dtype": torch.int64,
            },
        }
    )

    if gen == 8:
        buffer_specs.update(
            {
                "max_move_mask": {
                    "size": (trajectory_length, 4),
                    "dtype": torch.bool,
                },
                "max_move_policy": {
                    "size": (trajectory_length, 4),
                    "dtype": torch.float32,
                },
                "max_move_index": {
                    "size": (trajectory_length,),
                    "dtype": torch.int64,
                },
            }
        )

    buffer_specs.update(
        {
            "valid": {
                "size": (trajectory_length,),
                "dtype": torch.bool,
            }
        }
    )
    return buffer_specs


def create_buffers(num_buffers: int, trajectory_length: int, gen: int, gametype: str):
    """
    num_buffers: int
        the size of the replay buffer
    trajectory_length: int
        the max length of a replay
    battle_format: str
        the type of format
    """

    private_reserve_size = _get_private_reserve_size(gen)

    buffer_specs = get_buffer_specs(
        trajectory_length, gen, gametype, private_reserve_size
    )

    buffers: Dict[str, List[torch.Tensor]] = {key: [] for key in buffer_specs}
    for _ in range(num_buffers):
        for key in buffer_specs:
            if key.endswith("_mask"):
                buffers[key].append(torch.ones(**buffer_specs[key]).share_memory_())
            else:
                buffers[key].append(torch.zeros(**buffer_specs[key]).share_memory_())
    return buffers


def expand_bt(tensor: torch.Tensor, time: int = 1, batch: int = 1) -> torch.Tensor:
    shape = tensor.shape
    return tensor.view(time, batch, *(tensor.shape or shape))



from meloetta.frameworks.nash_ketchum.buffer import ReplayBuffer
from meloetta.frameworks.nash_ketchum.actor import NAshKetchumActor
from meloetta.frameworks.nash_ketchum.model import NAshKetchumModel
from meloetta.frameworks.nash_ketchum.learner import NAshKetchumLearner
from meloetta.frameworks.nash_ketchum.config import NAshKetchumConfig
from meloetta.frameworks.nash_ketchum import utils
